import base64
from collections.abc import Callable
from io import BytesIO

import pytest
from PIL import Image

from mistral_common.exceptions import InvalidRequestException, TokenizerException
from mistral_common.protocol.instruct.chunk import (
    AudioChunk,
    AudioURLChunk,
    ImageURLChunk,
    TextChunk,
    ThinkChunk,
)
from mistral_common.protocol.instruct.messages import (
    AssistantMessage,
    ChatMessage,
    SystemMessage,
    ToolMessage,
    UserMessage,
)
from mistral_common.protocol.instruct.normalize import get_normalizer
from mistral_common.protocol.instruct.request import (
    ChatCompletionRequest,
    InstructRequest,
    ModelSettings,
    ReasoningEffort,
)
from mistral_common.protocol.instruct.tool_calls import Function, FunctionCall, Tool, ToolCall
from mistral_common.protocol.instruct.validator import ValidationMode, get_validator
from mistral_common.tokens.tokenizers.audio import AudioConfig, AudioEncoder, AudioSpectrogramConfig, SpecialAudioIDs
from mistral_common.tokens.tokenizers.base import SpecialTokens, TokenizerVersion
from mistral_common.tokens.tokenizers.image import ImageConfig, ImageEncoder, SpecialImageIDs
from mistral_common.tokens.tokenizers.instruct import InstructTokenizerV15
from mistral_common.tokens.tokenizers.mistral import MistralTokenizer
from mistral_common.tokens.tokenizers.model_settings_builder import EnumBuilder, ModelSettingsBuilder
from mistral_common.tokens.tokenizers.tekken import Tekkenizer
from tests.fixtures.audio import get_dummy_audio_chunk, get_dummy_audio_url_chunk
from tests.test_tekken import get_special_tokens, quick_vocab

EXPECTED_TEXT_V15: str = (
    r"<s>[SYSTEM_PROMPT]S[/SYSTEM_PROMPT]"
    r'[AVAILABLE_TOOLS][{"type": "function", "function": {"name": "math_interpreter",'
    r' "description": "Get the value of an arithmetic expression.",'
    r' "parameters": {"type": "object", "properties": {"expression":'
    r' {"type": "string", "description": "Math expression."}}}}}]'
    r'[/AVAILABLE_TOOLS][MODEL_SETTINGS]{"reasoning_effort": "high"}[/MODEL_SETTINGS]'
    r"[INST]U1[/INST]A1"
    r"[TOOL_CALLS]F1[ARGS]{}[TOOL_CALLS]F2[ARGS]{}</s>"
    r"[TOOL_RESULTS]R1[/TOOL_RESULTS]"
    r"[TOOL_RESULTS]R2[/TOOL_RESULTS]A2</s>"
    r"[INST]U2[/INST]"
)

EXPECTED_TEXT_V15_NO_TOOLS: str = (
    r"<s>[SYSTEM_PROMPT]S[/SYSTEM_PROMPT]"
    r'[MODEL_SETTINGS]{"reasoning_effort": "high"}[/MODEL_SETTINGS]'
    r"[INST]U1[/INST]A1"
    r"[TOOL_CALLS]F1[ARGS]{}[TOOL_CALLS]F2[ARGS]{}</s>"
    r"[TOOL_RESULTS]R1[/TOOL_RESULTS]"
    r"[TOOL_RESULTS]R2[/TOOL_RESULTS]A2</s>"
    r"[INST]U2[/INST]"
)

EXPECTED_TEXT_TOOL_AUDIO: str = (
    r"<s>"
    r'[AVAILABLE_TOOLS][{"type": "function", "function": {"name": "fn",'
    r' "description": "test", "parameters": {}}}]'
    r'[/AVAILABLE_TOOLS][MODEL_SETTINGS]{"reasoning_effort": "none"}[/MODEL_SETTINGS]'
    r"[INST]Use the tool[/INST]"
    r"[TOOL_CALLS]fn[ARGS]{}</s>"
    r"[TOOL_RESULTS]result[BEGIN_AUDIO][AUDIO][AUDIO][/TOOL_RESULTS]"
)

EXPECTED_TEXT_TOOL_IMAGE: str = (
    r"<s>"
    r'[AVAILABLE_TOOLS][{"type": "function", "function": {"name": "fn",'
    r' "description": "test", "parameters": {}}}]'
    r'[/AVAILABLE_TOOLS][MODEL_SETTINGS]{"reasoning_effort": "none"}[/MODEL_SETTINGS]'
    r"[INST]Use the tool[/INST]"
    r"[TOOL_CALLS]fn[ARGS]{}</s>"
    r"[TOOL_RESULTS]result[IMG][IMG_END][/TOOL_RESULTS]"
)

EXPECTED_TEXT_SYSTEM_AUDIO: str = (
    r"<s>[SYSTEM_PROMPT]System with content[BEGIN_AUDIO][AUDIO][AUDIO][/SYSTEM_PROMPT]"
    r'[MODEL_SETTINGS]{"reasoning_effort": "none"}[/MODEL_SETTINGS]'
    r"[INST]Hello[/INST]"
)

EXPECTED_TEXT_USER_AUDIO: str = (
    r"<s>"
    r'[MODEL_SETTINGS]{"reasoning_effort": "none"}[/MODEL_SETTINGS]'
    r"[INST]Here is content[BEGIN_AUDIO][AUDIO][AUDIO][/INST]"
)

EXPECTED_TEXT_USER_IMAGE: str = (
    r"<s>"
    r'[MODEL_SETTINGS]{"reasoning_effort": "none"}[/MODEL_SETTINGS]'
    r"[INST][IMG][IMG_END]Here is content[/INST]"
)


def _build_v15_tekkenizer(model_settings_builder: ModelSettingsBuilder | None) -> Tekkenizer:
    r"""Build a v15 Tekkenizer with the given model settings builder.

    Args:
        model_settings_builder: The model settings builder, or None to create
            a tekkenizer without model settings support.
    """
    return Tekkenizer(
        quick_vocab([b"a", b"b", b"c", b"f", b"de"]),
        special_tokens=get_special_tokens(TokenizerVersion.v15, add_think=True),
        pattern=r".+",
        vocab_size=256 + 100,
        num_special_tokens=100,
        version=TokenizerVersion.v15,
        model_settings_builder=model_settings_builder,
    )


def get_v15_tekkenizer(
    model_settings_builder: ModelSettingsBuilder | None,
) -> InstructTokenizerV15:
    """Build an InstructTokenizerV15 with the given model settings builder."""
    return InstructTokenizerV15(_build_v15_tekkenizer(model_settings_builder))


def get_v15_mistral_tokenizer(
    model_settings_builder: ModelSettingsBuilder | None,
) -> MistralTokenizer:
    """Build a MistralTokenizer wrapping a v15 instruct tokenizer."""
    tekkenizer = _build_v15_tekkenizer(model_settings_builder)
    request_normalizer = get_normalizer(TokenizerVersion.v15, tekkenizer.model_settings_builder)
    validator = get_validator(TokenizerVersion.v15, mode=ValidationMode.test)
    return MistralTokenizer(
        InstructTokenizerV15(tekkenizer),
        validator=validator,
        request_normalizer=request_normalizer,
    )


def _build_model_settings_builder(
    allowed_reasoning_effort: tuple[str, ...] | None,
) -> ModelSettingsBuilder:
    """Build a ModelSettingsBuilder from allowed reasoning effort values.

    When `allowed_reasoning_effort` is `None`, returns `ModelSettingsBuilder.none()`
    (all fields ignored). This matches the behavior of `Tekkenizer.from_file` when no
    `model_settings_builder` key is present in the JSON.
    """
    if allowed_reasoning_effort is None:
        return ModelSettingsBuilder.none()
    return ModelSettingsBuilder(
        reasoning_effort=EnumBuilder[ReasoningEffort](
            values=[ReasoningEffort(v) for v in allowed_reasoning_effort],
            accepts_none=True,
            default=ReasoningEffort(allowed_reasoning_effort[0]) if allowed_reasoning_effort else None,
        )
    )


def _get_dummy_image_url_chunk() -> ImageURLChunk:
    r"""Build a small base64-encoded image URL chunk for testing."""
    img = Image.new("RGB", (4, 4), "red")
    buf = BytesIO()
    img.save(buf, "PNG")
    data_url = f"data:image/png;base64,{base64.b64encode(buf.getvalue()).decode()}"
    return ImageURLChunk(image_url=data_url)


def get_v15_mistral_tokenizer_with_audio() -> MistralTokenizer:
    r"""Build a V15 MistralTokenizer with audio encoder."""
    builder = _build_model_settings_builder(tuple(ReasoningEffort))
    tekkenizer = Tekkenizer(
        quick_vocab([b"a", b"b", b"c", b"f", b"de"]),
        special_tokens=get_special_tokens(TokenizerVersion.v15, add_audio=True),
        pattern=r".+",
        vocab_size=256 + 100,
        num_special_tokens=100,
        version=TokenizerVersion.v15,
        model_settings_builder=builder,
    )
    audio_config = AudioConfig(
        sampling_rate=24_000,
        frame_rate=12.5,
        encoding_config=AudioSpectrogramConfig(
            num_mel_bins=128,
            hop_length=160,
            window_size=400,
        ),
    )
    special_audio_ids = SpecialAudioIDs(
        audio=tekkenizer.get_special_token(SpecialTokens.audio.value),
        begin_audio=tekkenizer.get_special_token(SpecialTokens.begin_audio.value),
        streaming_pad=None,
        text_to_audio=None,
        audio_to_text=None,
    )
    audio_encoder = AudioEncoder(audio_config, special_audio_ids)
    instruct_tokenizer = InstructTokenizerV15(tekkenizer, audio_encoder=audio_encoder)
    request_normalizer = get_normalizer(TokenizerVersion.v15, tekkenizer.model_settings_builder)
    validator = get_validator(TokenizerVersion.v15, mode=ValidationMode.test)
    return MistralTokenizer(
        instruct_tokenizer=instruct_tokenizer,
        validator=validator,
        request_normalizer=request_normalizer,
    )


def get_v15_mistral_tokenizer_with_image() -> MistralTokenizer:
    r"""Build a V15 MistralTokenizer with image encoder."""
    builder = _build_model_settings_builder(tuple(ReasoningEffort))
    tekkenizer = Tekkenizer(
        quick_vocab([b"a", b"b", b"c", b"f", b"de"]),
        special_tokens=get_special_tokens(TokenizerVersion.v15, add_think=True),
        pattern=r".+",
        vocab_size=256 + 100,
        num_special_tokens=100,
        version=TokenizerVersion.v15,
        model_settings_builder=builder,
    )
    image_config = ImageConfig(image_patch_size=16, max_image_size=1024)
    special_image_ids = SpecialImageIDs(
        img=tekkenizer.get_special_token(SpecialTokens.img.value),
        img_break=tekkenizer.get_special_token(SpecialTokens.img_break.value),
        img_end=tekkenizer.get_special_token(SpecialTokens.img_end.value),
    )
    image_encoder = ImageEncoder(image_config, special_image_ids)
    instruct_tokenizer = InstructTokenizerV15(tekkenizer, image_encoder=image_encoder)
    request_normalizer = get_normalizer(TokenizerVersion.v15, tekkenizer.model_settings_builder)
    validator = get_validator(TokenizerVersion.v15, mode=ValidationMode.test)
    return MistralTokenizer(
        instruct_tokenizer=instruct_tokenizer,
        validator=validator,
        request_normalizer=request_normalizer,
    )


@pytest.fixture(scope="session")
def v15_tekkenizer() -> InstructTokenizerV15:
    return get_v15_tekkenizer(_build_model_settings_builder(tuple(ReasoningEffort)))


@pytest.fixture(scope="session")
def v15_tekkenizer_no_reasoning() -> InstructTokenizerV15:
    return get_v15_tekkenizer(_build_model_settings_builder(None))


@pytest.fixture
def available_tools() -> list[Tool]:
    return [
        Tool(
            function=Function(
                name="math_interpreter",
                description="Get the value of an arithmetic expression.",
                parameters={
                    "type": "object",
                    "properties": {
                        "expression": {
                            "type": "string",
                            "description": "Math expression.",
                        }
                    },
                },
            )
        )
    ]


@pytest.fixture
def messages() -> list[ChatMessage]:
    return [
        SystemMessage(content="S"),
        UserMessage(content="U1"),
        AssistantMessage(
            content="A1",
            tool_calls=[
                ToolCall(id="123456789", function=FunctionCall(name="F1", arguments="{}")),
                ToolCall(id="999999999", function=FunctionCall(name="F2", arguments="{}")),
            ],
        ),
        ToolMessage(content="R1", tool_call_id="123456789"),
        ToolMessage(content="R2", tool_call_id="999999999"),
        AssistantMessage(content="A2"),
        UserMessage(content="U2"),
    ]


@pytest.fixture(scope="session")
def audio_chunk() -> AudioChunk:
    return get_dummy_audio_chunk()


# Multimodal content chunks and their corresponding tokenizer factories for parametrized tests.
_TOOL_MULTIMODAL_PARAMS = [
    pytest.param(
        get_dummy_audio_chunk(), 1, 0, get_v15_mistral_tokenizer_with_audio, EXPECTED_TEXT_TOOL_AUDIO, id="audio"
    ),
    pytest.param(
        get_dummy_audio_url_chunk(),
        1,
        0,
        get_v15_mistral_tokenizer_with_audio,
        EXPECTED_TEXT_TOOL_AUDIO,
        id="audio_url",
    ),
    pytest.param(
        _get_dummy_image_url_chunk(),
        0,
        1,
        get_v15_mistral_tokenizer_with_image,
        EXPECTED_TEXT_TOOL_IMAGE,
        id="image_url",
    ),
]
_SYSTEM_MULTIMODAL_PARAMS = [
    pytest.param(
        get_dummy_audio_chunk(), 1, 0, get_v15_mistral_tokenizer_with_audio, EXPECTED_TEXT_SYSTEM_AUDIO, id="audio"
    ),
]
_USER_MULTIMODAL_PARAMS = [
    pytest.param(
        get_dummy_audio_chunk(), 1, 0, get_v15_mistral_tokenizer_with_audio, EXPECTED_TEXT_USER_AUDIO, id="audio"
    ),
    pytest.param(
        get_dummy_audio_url_chunk(),
        1,
        0,
        get_v15_mistral_tokenizer_with_audio,
        EXPECTED_TEXT_USER_AUDIO,
        id="audio_url",
    ),
    pytest.param(
        _get_dummy_image_url_chunk(),
        0,
        1,
        get_v15_mistral_tokenizer_with_image,
        EXPECTED_TEXT_USER_IMAGE,
        id="image_url",
    ),
]


def test_tools_and_reasoning_effort(
    v15_tekkenizer: InstructTokenizerV15, available_tools: list[Tool], messages: list[ChatMessage]
) -> None:
    request = InstructRequest(
        messages=messages,
        available_tools=available_tools,
        settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
    )
    tokenized = v15_tekkenizer.encode_instruct(request)
    assert tokenized.text == EXPECTED_TEXT_V15, tokenized.text


def test_no_tools_and_reasoning_effort(v15_tekkenizer: InstructTokenizerV15, messages: list[ChatMessage]) -> None:
    request: InstructRequest = InstructRequest(
        messages=messages, available_tools=None, settings=ModelSettings(reasoning_effort=ReasoningEffort.none)
    )
    tokenized = v15_tekkenizer.encode_instruct(request)
    expected_text_no_tools = EXPECTED_TEXT_V15_NO_TOOLS.replace("high", "none")
    assert tokenized.text == expected_text_no_tools, tokenized.text


def test_no_settings_does_not_encode_model_settings(
    v15_tekkenizer_no_reasoning: InstructTokenizerV15, messages: list[ChatMessage]
) -> None:
    request: InstructRequest = InstructRequest(messages=messages, available_tools=None, settings=ModelSettings.none())
    tokenized = v15_tekkenizer_no_reasoning.encode_instruct(request)
    assert "[MODEL_SETTINGS]" not in (tokenized.text or "")


def test_system_think_chunk_raises_v15(v15_tekkenizer: InstructTokenizerV15) -> None:
    messages = [SystemMessage(content=[ThinkChunk(thinking="Hi")])]
    request: InstructRequest = InstructRequest(
        messages=messages, settings=ModelSettings(reasoning_effort=ReasoningEffort.high)
    )
    with pytest.raises(TokenizerException, match="ThinkChunk in system message is not supported for this model"):
        v15_tekkenizer.encode_instruct(request)


@pytest.mark.parametrize(
    ("reasoning_effort", "allowed_reasoning_effort", "raises", "match"),
    [
        (None, ("none", "high"), None, None),
        ("none", ("none", "high"), None, None),
        ("high", ("none", "high"), None, None),
        ("high", ("none",), InvalidRequestException, "should be one of"),
        ("none", ("none",), None, None),
        ("none", (), InvalidRequestException, "not supported for this model"),
    ],
)
def test_forbidden_reasoning_effort_raises(
    available_tools: list[Tool],
    messages: list[ChatMessage],
    reasoning_effort: str | None,
    allowed_reasoning_effort: tuple[str, ...] | None,
    raises: type[Exception] | None,
    match: str | None,
) -> None:
    builder = _build_model_settings_builder(allowed_reasoning_effort)
    tokenizer_v15 = get_v15_mistral_tokenizer(builder)
    request = ChatCompletionRequest(
        messages=messages,
        tools=available_tools,
        reasoning_effort=reasoning_effort,  # type: ignore[arg-type]
    )
    if raises is not None:
        assert match is not None
        with pytest.raises(raises, match=match):
            tokenizer_v15.encode_chat_completion(request)
    else:
        assert match is None
        tokenizer_v15.encode_chat_completion(request)


@pytest.mark.parametrize(("reasoning_effort", "allowed_reasoning_effort"), [(None, ()), ("none", None)])
def test_encode_ignore_one_model_settings(
    messages: list[ChatMessage],
    reasoning_effort: str | None,
    allowed_reasoning_effort: tuple[str, ...] | None,
) -> None:
    builder = _build_model_settings_builder(allowed_reasoning_effort)
    tokenizer_v15 = get_v15_mistral_tokenizer(builder)
    request = ChatCompletionRequest(messages=messages, reasoning_effort=reasoning_effort)  # type: ignore[arg-type]
    tokenized = tokenizer_v15.encode_chat_completion(request)
    assert "[MODEL_SETTINGS]" not in (tokenized.text or "")


@pytest.mark.parametrize("reasoning_effort", [None, *list(ReasoningEffort)])
def test_end_to_end_with_default(messages: list[ChatMessage], reasoning_effort: ReasoningEffort | None) -> None:
    """When the builder has a default, None reasoning_effort resolves to the default."""
    builder = _build_model_settings_builder(("none", "high"))
    tokenizer_v15 = get_v15_mistral_tokenizer(builder)
    request = ChatCompletionRequest(messages=messages, reasoning_effort=reasoning_effort)
    tokenized = tokenizer_v15.encode_chat_completion(request)
    text = tokenized.text or ""
    if reasoning_effort == ReasoningEffort.high:
        assert '[MODEL_SETTINGS]{"reasoning_effort": "high"}[/MODEL_SETTINGS]' in text
    else:
        # Both None (default applied) and explicit "none" resolve to the default "none"
        assert '[MODEL_SETTINGS]{"reasoning_effort": "none"}[/MODEL_SETTINGS]' in text


@pytest.mark.parametrize("reasoning_effort", [None, *list(ReasoningEffort)])
def test_end_to_end_no_default(messages: list[ChatMessage], reasoning_effort: ReasoningEffort | None) -> None:
    """When the builder has no default, None reasoning_effort means no model settings."""
    builder = ModelSettingsBuilder(
        reasoning_effort=EnumBuilder[ReasoningEffort](values=list(ReasoningEffort), accepts_none=True, default=None)
    )
    tokenizer_v15 = get_v15_mistral_tokenizer(builder)
    request = ChatCompletionRequest(messages=messages, reasoning_effort=reasoning_effort)
    tokenized = tokenizer_v15.encode_chat_completion(request)
    text = tokenized.text or ""
    if reasoning_effort == ReasoningEffort.high:
        assert '[MODEL_SETTINGS]{"reasoning_effort": "high"}[/MODEL_SETTINGS]' in text
    elif reasoning_effort == ReasoningEffort.none:
        assert '[MODEL_SETTINGS]{"reasoning_effort": "none"}[/MODEL_SETTINGS]' in text
    else:
        # None with no default -> no model settings encoded
        assert "[MODEL_SETTINGS]" not in text


def test_encode_chat_completion_continue_final_message() -> None:
    builder = _build_model_settings_builder(("none", "high"))
    tokenizer_v15 = get_v15_mistral_tokenizer(builder)
    request: ChatCompletionRequest = ChatCompletionRequest(
        messages=[UserMessage(content="a"), AssistantMessage(content="b")],
        continue_final_message=True,
    )
    encoded = tokenizer_v15.encode_chat_completion(request)

    eos_id = tokenizer_v15.instruct_tokenizer.tokenizer.eos_id
    assert encoded.tokens[-1] != eos_id


@pytest.mark.parametrize(
    ("content_chunk", "expected_audios", "expected_images", "tokenizer_factory", "expected_text"),
    _TOOL_MULTIMODAL_PARAMS,
)
def test_encode_chat_completion_with_multimodal_tool(
    content_chunk: AudioChunk | AudioURLChunk | ImageURLChunk,
    expected_audios: int,
    expected_images: int,
    tokenizer_factory: Callable[[], MistralTokenizer],
    expected_text: str,
) -> None:
    mistral_tokenizer = tokenizer_factory()
    chat_request: ChatCompletionRequest = ChatCompletionRequest(
        messages=[
            UserMessage(content="Use the tool"),
            AssistantMessage(tool_calls=[ToolCall(id="test12345", function=FunctionCall(name="fn", arguments="{}"))]),
            ToolMessage(
                content=[TextChunk(text="result"), content_chunk],
                tool_call_id="test12345",
            ),
        ],
        tools=[Tool(function=Function(name="fn", description="test", parameters={}))],
    )
    encoded = mistral_tokenizer.encode_chat_completion(chat_request)
    assert encoded.text == expected_text, encoded.text
    assert len(encoded.audios) == expected_audios
    assert len(encoded.images) == expected_images


@pytest.mark.parametrize(
    ("content_chunk", "expected_audios", "expected_images", "tokenizer_factory", "expected_text"),
    _SYSTEM_MULTIMODAL_PARAMS,
)
def test_encode_chat_completion_with_multimodal_system(
    content_chunk: AudioChunk,
    expected_audios: int,
    expected_images: int,
    tokenizer_factory: Callable[[], MistralTokenizer],
    expected_text: str,
) -> None:
    mistral_tokenizer = tokenizer_factory()
    chat_request: ChatCompletionRequest = ChatCompletionRequest(
        messages=[
            SystemMessage(content=[TextChunk(text="System with content"), content_chunk]),
            UserMessage(content="Hello"),
        ],
    )
    encoded = mistral_tokenizer.encode_chat_completion(chat_request)
    assert encoded.text == expected_text, encoded.text
    assert len(encoded.audios) == expected_audios
    assert len(encoded.images) == expected_images


@pytest.mark.parametrize(
    ("content_chunk", "expected_audios", "expected_images", "tokenizer_factory", "expected_text"),
    _USER_MULTIMODAL_PARAMS,
)
def test_encode_chat_completion_with_multimodal_user(
    content_chunk: AudioChunk | AudioURLChunk | ImageURLChunk,
    expected_audios: int,
    expected_images: int,
    tokenizer_factory: Callable[[], MistralTokenizer],
    expected_text: str,
) -> None:
    mistral_tokenizer = tokenizer_factory()
    chat_request = ChatCompletionRequest(
        messages=[
            UserMessage(content=[TextChunk(text="Here is content"), content_chunk]),
        ],
    )
    encoded = mistral_tokenizer.encode_chat_completion(chat_request)
    assert encoded.text == expected_text, encoded.text
    assert len(encoded.audios) == expected_audios
    assert len(encoded.images) == expected_images



import pytest

from mistral_common.exceptions import (
    InvalidToolException,
    InvalidToolSchemaException,
)
from mistral_common.protocol.instruct.tool_calls import Function, Tool
from mistral_common.protocol.instruct.validator import (
    MistralRequestValidator,
    MistralRequestValidatorV3,
    MistralRequestValidatorV5,
    MistralRequestValidatorV13,
    MistralRequestValidatorV15,
    ValidationMode,
    get_validator,
)
from mistral_common.tokens.tokenizers.base import TokenizerVersion


class TestValidateTools:
    @pytest.fixture
    def validator(self) -> MistralRequestValidator:
        return MistralRequestValidator()

    def test_tool_function_bad_name(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(InvalidToolException, match=r"must be a-z, A-Z, 0-9"):
            validator._validate_tools(
                tools=[
                    Tool(
                        function=Function(
                            name=')_(_)~@:}>""',
                            parameters={
                                "type": "object",
                                "properties": {
                                    "location": {
                                        "type": "string",
                                        "description": "The city and state, e.g. San Francisco, CA",
                                    },
                                },
                            },
                        )
                    )
                ]
            )

    def test_tool_function_invalid_schema(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(InvalidToolSchemaException, match=r"32 is not valid under any of the given schemas"):
            validator._validate_tools(tools=[Tool(function=Function(name="function_name", parameters={"type": 32}))])


@pytest.mark.parametrize(
    "version,expected_class",
    [
        (TokenizerVersion.v1, MistralRequestValidator),
        (TokenizerVersion.v2, MistralRequestValidator),
        (TokenizerVersion.v3, MistralRequestValidatorV3),
        (TokenizerVersion.v7, MistralRequestValidatorV5),
        (TokenizerVersion.v13, MistralRequestValidatorV13),
        (TokenizerVersion.v15, MistralRequestValidatorV15),
    ],
)
def test_get_validator_version_mapping(version: TokenizerVersion, expected_class: type) -> None:
    validator = get_validator(version, ValidationMode.test)
    assert isinstance(validator, expected_class)
    assert validator._mode == ValidationMode.test


@pytest.mark.parametrize("mode", list(ValidationMode))
def test_get_validator_mode_passed_correctly(mode: ValidationMode) -> None:
    validator = get_validator(TokenizerVersion.v2, mode)
    assert validator._mode == validator.mode == mode



import os.path
import warnings

import pytest

import mistral_common.deprecation
from mistral_common.deprecation import deprecated_import, warn_once


@pytest.fixture(autouse=True)
def _clear_warned_keys() -> None:
    mistral_common.deprecation._warned_keys.clear()


def test_deprecated_import_returns_correct_object() -> None:
    result = deprecated_import("old.module", "os.path", "join", "99.0")
    assert result is os.path.join


def test_deprecated_import_emits_deprecation_warning() -> None:
    with pytest.warns(DeprecationWarning, match=r"Importing join from old\.mod") as record:
        deprecated_import("old.mod", "os.path", "join", "99.0")

    assert len(record) == 1
    msg = str(record[0].message)
    assert "Use os.path.join instead" in msg
    assert "Will be removed in 99.0" in msg


def test_deprecated_import_warns_only_once() -> None:
    with pytest.warns(DeprecationWarning):
        deprecated_import("old.mod", "os.path", "join", "99.0")

    with warnings.catch_warnings(record=True) as caught:
        warnings.simplefilter("always")
        deprecated_import("old.mod", "os.path", "join", "99.0")

    dep_warnings = [w for w in caught if issubclass(w.category, DeprecationWarning)]
    assert dep_warnings == []


def test_deprecated_import_different_pairs_each_warn() -> None:
    with pytest.warns(DeprecationWarning, match="join"):
        deprecated_import("pkg.a", "os.path", "join", "1.0")

    with pytest.warns(DeprecationWarning, match="exists"):
        deprecated_import("pkg.b", "os.path", "exists", "1.0")


def test_deprecated_import_raises_attribute_error() -> None:
    with pytest.raises(AttributeError):
        deprecated_import("old", "os.path", "no_such_attr_xyz", "1.0")


def test_deprecated_import_raises_module_not_found_error() -> None:
    with pytest.raises(ModuleNotFoundError):
        deprecated_import("old", "no_such_module_xyz_abc", "Foo", "1.0")


def test_warn_once_emits_warning() -> None:
    with pytest.warns(DeprecationWarning, match="something broke"):
        warn_once("k1", "something broke", DeprecationWarning, stacklevel=2)


def test_warn_once_does_not_repeat() -> None:
    with pytest.warns(DeprecationWarning):
        warn_once("k1", "msg", DeprecationWarning, stacklevel=2)

    with warnings.catch_warnings(record=True) as caught:
        warnings.simplefilter("always")
        warn_once("k1", "msg", DeprecationWarning, stacklevel=2)

    dep_warnings = [w for w in caught if issubclass(w.category, DeprecationWarning)]
    assert dep_warnings == []


def test_warn_once_different_keys_each_warn() -> None:
    with pytest.warns(DeprecationWarning, match="first"):
        warn_once("a", "first", DeprecationWarning, stacklevel=2)

    with pytest.warns(DeprecationWarning, match="second"):
        warn_once("b", "second", DeprecationWarning, stacklevel=2)



import warnings
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Any, Generic, TypeVar

import numpy as np
from pydantic import ConfigDict, Field

from mistral_common.base import MistralBase
from mistral_common.protocol.fim.request import FIMRequest
from mistral_common.protocol.instruct.chunk import ContentChunk
from mistral_common.protocol.instruct.messages import (
    AssistantMessageType,
    UserMessage,
)
from mistral_common.protocol.instruct.request import InstructRequest
from mistral_common.protocol.instruct.tool_calls import Tool
from mistral_common.protocol.speech.request import SpeechRequest
from mistral_common.protocol.transcription.request import TranscriptionRequest
from mistral_common.tokens.tokenizers.audio import Audio, AudioEncoder
from mistral_common.tokens.tokenizers.image import ImageEncoder
from mistral_common.tokens.tokenizers.model_settings_builder import ModelSettingsBuilder


class UserMessagePosition(str, Enum):
    """Where to encode available tools"""

    first = "first"
    last = "last"


class SpecialTokens(str, Enum):
    r"""Enum of special tokens used in the tokenizer.

    Attributes:
        unk: The unknown token.
        bos: The beginning of string token.
        eos: The end of string token.
        begin_inst: The beginning of instruction token.
        end_inst: The end of instruction token.
        begin_tools: The beginning of tools token.
        end_tools: The end of tools token.
        begin_tool_results: The beginning of tool results token.
        end_tool_results: The end of tool results token.
        tool_calls: The tool calls token.
        img: The image token.
        pad: The pad token.
        img_break: The image break token.
        img_end: The image end token.
        prefix: The prefix token for FIM.
        middle: The middle token for FIM.
        suffix: The suffix token for FIM.
        begin_system: The beginning of system prompt token.
        end_system: The end of system prompt token.
        begin_tool_content: The beginning of tool content token.
        args: The args token.
        call_id: The call id token.
        audio: The audio token.
        begin_audio: The beginning of audio token.
        transcribe: The transcribe token.
        begin_think: The beginning of think token.
        end_think: The end of think token.
        streaming_pad: The streaming pad token.
        streaming_word: The streaming word token.
        text_to_audio: The text to audio token.
        audio_to_text: The audio to text token.

    Examples:
        >>> unk = SpecialTokens.unk
    """

    unk = "<unk>"
    bos = "<s>"
    eos = "</s>"
    begin_inst = "[INST]"
    end_inst = "[/INST]"
    begin_tools = "[AVAILABLE_TOOLS]"
    end_tools = "[/AVAILABLE_TOOLS]"
    begin_tool_results = "[TOOL_RESULTS]"
    end_tool_results = "[/TOOL_RESULTS]"
    tool_calls = "[TOOL_CALLS]"
    img = "[IMG]"
    pad = "<pad>"
    img_break = "[IMG_BREAK]"
    img_end = "[IMG_END]"
    prefix = "[PREFIX]"
    middle = "[MIDDLE]"
    suffix = "[SUFFIX]"
    begin_system = "[SYSTEM_PROMPT]"
    end_system = "[/SYSTEM_PROMPT]"
    begin_tool_content = "[TOOL_CONTENT]"
    args = "[ARGS]"
    call_id = "[CALL_ID]"
    audio = "[AUDIO]"
    begin_audio = "[BEGIN_AUDIO]"
    transcribe = "[TRANSCRIBE]"
    begin_think = "[THINK]"
    end_think = "[/THINK]"
    streaming_pad = "[STREAMING_PAD]"
    streaming_word = "[STREAMING_WORD]"
    text_to_audio = "[NEXT_AUDIO_TEXT]"
    audio_to_text = "[REPEAT_AUDIO_TEXT]"
    begin_model_settings = "[MODEL_SETTINGS]"
    end_model_settings = "[/MODEL_SETTINGS]"


class SpecialTokenPolicy(str, Enum):
    r"""What to do with special tokens when encoding/decoding.

    Attributes:
        IGNORE: Ignore special tokens.
        KEEP: Keep special tokens.
        RAISE: Raise an error if special tokens are found.
    """

    IGNORE = "ignore"
    KEEP = "keep"
    RAISE = "raise"

    @classmethod
    def _missing_(cls, value: Any) -> Any:
        # Backward compatibility of int values.
        match value:
            case 0:
                return SpecialTokenPolicy.IGNORE
            case 1:
                return SpecialTokenPolicy.KEEP
            case 2:
                return SpecialTokenPolicy.RAISE
        return super()._missing_(value)


class TokenizerVersion(str, Enum):
    r"""Enum of tokenizer versions.

    Allow to distinguish between different versions of the tokenizer and maintain backward compatibility.

    Attributes:
        v1: The first version of the tokenizer.
        v2: The second version of the tokenizer that includes special control tokens [INST], [\INST].
        v3: The third version of the tokenizer that includes improved function calling.
        v7: The seventh version of the tokenizer that includes improved system prompt and function calling.
        v11: The eleventh version of the tokenizer that includes improved function calling.
        v13: The thirteenth version of the tokenizer that includes no call id tokenization and better prompt caching.

    Examples:
        >>> version = TokenizerVersion.v1
    """

    @property
    def _version_num(self) -> int:
        warnings.warn(
            "`TokenizerVersion._version_num` is deprecated in favor to `TokenizerVersion.version_num` and will "
            "be removed in 1.13.0."
        )
        return self.version_num

    @property
    def version_num(self) -> int:
        return int(self.value[1:])

    @property
    def supports_model_settings(self) -> bool:
        return self >= TokenizerVersion.v15

    def __lt__(self, other: "str | TokenizerVersion") -> bool:
        if isinstance(other, str):
            other = TokenizerVersion(other)
        return self.version_num < other.version_num

    def __le__(self, other: "str | TokenizerVersion") -> bool:
        if isinstance(other, str):
            other = TokenizerVersion(other)
        return self.version_num <= other.version_num

    def __gt__(self, other: "str | TokenizerVersion") -> bool:
        if isinstance(other, str):
            other = TokenizerVersion(other)
        return self.version_num > other.version_num

    def __ge__(self, other: "str | TokenizerVersion") -> bool:
        if isinstance(other, str):
            other = TokenizerVersion(other)
        return self.version_num >= other.version_num

    v1 = "v1"  # vocab_size = 32000
    v2 = "v2"  # vocab_size = 32768 with special control tokens [INST], [\INST]
    v3 = "v3"  # vocab_size = 32768 (spm) OR 131072 (tekken) with improved function calling
    v7 = "v7"  # vocab_size = 32768 (spm) or 131072 (tekken) with improved system prompt and function calling
    v11 = "v11"  # 131072 (tekken) with improved function calling
    v13 = "v13"  # 131072 (tekken) with no call id and better prompt caching
    v15 = "v15"  # 131072 (tekken) with model settings


class Tokenized(MistralBase):
    r"""A tokenized [`InstructRequest`][mistral_common.protocol.instruct.request.InstructRequest].

    Attributes:
        tokens: The token ids.
        text: The text representation of the tokens.
        prefix_ids: The prefix ids for FIM.
        images: The loaded images associated with the tokens.

    Examples:
        >>> tokenized = Tokenized(tokens=[1, 2, 3], text="Hello world", prefix_ids=[1], images=[])
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    tokens: list[int]
    text: str | None = None
    prefix_ids: list[int] | None = None
    images: list[np.ndarray] = Field(default_factory=list)
    audios: list[Audio] = Field(default_factory=list)


class Tokenizer(ABC):
    @property
    @abstractmethod
    def n_words(self) -> int:
        r"""Vocabulary size of the tokenizer."""

    @property
    @abstractmethod
    def special_ids(self) -> set[int]:
        r"""Ids of the special tokens."""

    @property
    @abstractmethod
    def num_special_tokens(self) -> int:
        r"""The number of special tokens of the tokenizer."""

    @property
    @abstractmethod
    def model_settings_builder(self) -> ModelSettingsBuilder | None:
        r"""The model settings builder, or None if unsupported by this version."""

    @abstractmethod
    def vocab(self) -> list[str]:
        r"""All tokens in the vocabulary as strings."""

    @abstractmethod
    def id_to_piece(self, token_id: int) -> str:
        r"""Convert a token id to the token str."""

    @property
    @abstractmethod
    def bos_id(self) -> int:
        r"""id of the Beginning of String token."""

    @property
    @abstractmethod
    def eos_id(self) -> int:
        r"""id of the End of String token."""

    @property
    @abstractmethod
    def pad_id(self) -> int:
        r"""id of the Pad token."""

    @property
    @abstractmethod
    def unk_id(self) -> int:
        r"""id of the Unk token."""

    @abstractmethod
    def encode(self, s: str, bos: bool, eos: bool) -> list[int]:
        """Convert a string to a list of token ids."""

    @abstractmethod
    def decode(self, tokens: list[int], special_token_policy: SpecialTokenPolicy = SpecialTokenPolicy.IGNORE) -> str:
        r"""Decode the token ids to a string.

        Args:
            tokens: The token ids to decode.
            special_token_policy: The policy to use for special tokens.

        Returns:
            The decoded string.
        """

    @abstractmethod
    def get_special_token(self, s: str) -> int:
        r"""Get the id of a control token."""

    @abstractmethod
    def is_special(self, token: int | np.integer | str) -> bool:
        r"""Check if token id or token str is a special token."""

    @property
    @abstractmethod
    def version(self) -> TokenizerVersion:
        r"""Get the version of the tokenizer."""

    @abstractmethod
    def _to_string(self, tokens: list[int]) -> str: ...

    @property
    @abstractmethod
    def file_path(self) -> Path:
        r"""The file path of the tokenizer."""
        ...


InstructRequestType = TypeVar("InstructRequestType", bound=InstructRequest)
FIMRequestType = TypeVar("FIMRequestType", bound=FIMRequest)
TokenizedType = TypeVar("TokenizedType", bound=Tokenized)


class InstructTokenizer(Generic[InstructRequestType, FIMRequestType, TokenizedType, AssistantMessageType]):
    r"""Base class for instruct tokenizers.

    Attributes:
        tokenizer: The tokenizer to use.
        image_encoder: The image encoder to use if any.
        audio_encoder: The audio encoder to use if any.
    """

    tokenizer: Tokenizer
    image_encoder: ImageEncoder | None
    audio_encoder: AudioEncoder | None

    @property
    def version(self) -> TokenizerVersion:
        r"""The version of the tokenizer."""
        return self.tokenizer.version

    def __init__(
        self, tokenizer: Tokenizer, image_encoder: ImageEncoder | None, audio_encoder: AudioEncoder | None
    ) -> None:
        r"""Initialize the instruct tokenizer.

        Args:
            tokenizer: The tokenizer to use.
            image_encoder: The image encoder to use if any.
            audio_encoder: The audio encoder to use if any.
        """

    @abstractmethod
    def encode_instruct(self, request: InstructRequestType) -> TokenizedType:
        r"""Instruct request to Tokenized object

        Args:
            request: The instruct request to encode.

        Returns:
            The tokenized instruct request.
        """

    @abstractmethod
    def encode_transcription(self, request: TranscriptionRequest) -> TokenizedType:
        r"""
        Encodes an audio transcription request into a tokenized format.

        This method processes a transcription request containing audio data,
        encodes the user message, and returns the tokenized output.

        Args:
            request: The transcription request object containing
                the audio data to be encoded.

        Returns:
            Tokenized: The tokenized representation of the audio data, including processed audio and tokens
        """

    @abstractmethod
    def encode_speech_request(self, request: SpeechRequest) -> TokenizedType:
        r"""Encodes a speech synthesis request into a tokenized format.

        This method processes a speech request containing text input and
        optional reference audio or voice preset, and returns the tokenized output.

        Args:
            request: The speech request object containing the text and voice/audio data.

        Returns:
            Tokenized: The tokenized representation of the speech request.
        """

    @abstractmethod
    def decode(self, tokens: list[int], special_token_policy: SpecialTokenPolicy) -> str:
        r"""Convert token ids to string

        Args:
            tokens: The token ids to decode.
            special_token_policy: The policy to use for special tokens.

        Returns:
            The decoded string.
        """

    @abstractmethod
    def encode_fim(self, request: FIMRequestType) -> TokenizedType:
        r"""FIM request to Tokenized object

        Args:
            request: The FIM request to encode.

        Returns:
            The tokenized FIM request.
        """

    @abstractmethod
    def encode_user_message(
        self,
        message: UserMessage,
        available_tools: list[Tool] | None,
        is_last: bool,
        is_first: bool,
        system_prompt: str | None = None,
        force_img_first: bool = False,
    ) -> tuple[list[int], list[np.ndarray], list[Audio]]:
        r"""Encode a user message.

        Args:
            message: The user message to encode.
            available_tools: The available tools.
            is_last: Whether the message is the last one.
            is_first: Whether the message is the first one.
            system_prompt: The system prompt.
            force_img_first: Whether to force the image to be first.

        Returns:
            The encoded tokens and images.
        """
        ...

    @abstractmethod
    def encode_user_content(
        self,
        content: str | list[ContentChunk],
        is_last: bool,
        system_prompt: str | None = None,
        force_img_first: bool = False,
    ) -> tuple[list[int], list[np.ndarray], list[Audio]]:
        r"""Encode a user content.

        Args:
            content: The user content to encode.
            is_last: Whether the content is the last one.
            system_prompt: The system prompt.
            force_img_first: Whether to force the image to be first.

        Returns:
            The encoded tokens and images.
        """
        ...

    @abstractmethod
    def _to_string(self, tokens: list[int]) -> str: ...



import pytest
from pydantic import ValidationError

from mistral_common.protocol.instruct.messages import UserMessage
from mistral_common.protocol.instruct.request import ChatCompletionRequest


class TestValidateRequest:
    @pytest.fixture
    def chat_request_raw(self) -> dict:
        return {"model": "test-model", "message": [UserMessage(content="foo")]}

    def test_request_random_seed_negative(self, chat_request_raw: dict) -> None:
        with pytest.raises(ValidationError):
            ChatCompletionRequest(**chat_request_raw, random_seed=-1)

    def test_from_openai_preserves_zero_seed(self) -> None:
        request = ChatCompletionRequest.from_openai(
            messages=[{"role": "user", "content": "hello"}],
            seed=0,
        )

        assert request.random_seed == 0



import pytest

from mistral_common.experimental.think import _split_content_and_think_chunks


@pytest.mark.parametrize(
    "tokens, expected",
    [
        ([], []),
        ([1, 2, 3, 4, 5], [([1, 2, 3, 4, 5], False)]),
        ([1, 2, 3, 100, 4, 5, 200, 6, 7, 8], [([1, 2, 3], False), ([100, 4, 5, 200], True), ([6, 7, 8], False)]),
        (
            [100, 1, 2, 3, 200, 4, 5, 6, 100, 7, 8, 9, 200],
            [([100, 1, 2, 3, 200], True), ([4, 5, 6], False), ([100, 7, 8, 9, 200], True)],
        ),
        ([100, 1, 2, 3, 200, 100, 4, 5, 6, 200], [([100, 1, 2, 3, 200], True), ([100, 4, 5, 6, 200], True)]),
        ([100, 1, 2, 3, 200, 100, 5, 6, 7], [([100, 1, 2, 3, 200], True), ([100, 5, 6, 7], True)]),
    ],
)
def test_split_content_and_think_chunks(tokens: list[int], expected: list[tuple[list[int], bool]]) -> None:
    content_or_think_chunks = _split_content_and_think_chunks(tokens, 100, 200)
    assert content_or_think_chunks == expected


@pytest.mark.parametrize(
    "tokens, error_message",
    [
        ([100, 1, 2, 3, 100, 200, 100, 5, 6, 7], r"Nested think chunks are not allowed."),
        ([100, 1, 2, 3, 200, 200, 5, 6, 7, 100], r"End think token found without a begin think token."),
    ],
)
def test_split_content_and_think_chunks_error(tokens: list[int], error_message: str) -> None:
    with pytest.raises(ValueError, match=error_message):
        _split_content_and_think_chunks(tokens, 100, 200)



from enum import Enum
from typing import Any, Generic

from pydantic import Field

from mistral_common.base import MistralBase
from mistral_common.protocol.base import BaseCompletionRequest
from mistral_common.protocol.instruct.converters import (
    convert_openai_messages,
    convert_openai_tools,
)
from mistral_common.protocol.instruct.messages import (
    AssistantMessage,
    ChatMessage,
    ChatMessageType,
    ReasoningFieldFormat,
)
from mistral_common.protocol.instruct.tool_calls import Tool, ToolChoice, ToolChoiceEnum, ToolType


class ResponseFormats(str, Enum):
    r"""Enum of the different formats of an instruct response.

    Attributes:
        text: The response is a plain text.
        json: The response is a JSON object.

    Examples:
        >>> response_format = ResponseFormats.text
    """

    text = "text"
    json = "json_object"


class ReasoningEffort(str, Enum):
    r"""Controls how much reasoning effort the model should apply.

    Attributes:
        none: No additional reasoning effort.
        high: High reasoning effort for complex tasks.
    """

    none = "none"
    high = "high"


class ModelSettings(MistralBase):
    r"""Model configuration settings for instruct requests.

    This class encapsulates various model configuration options that can be
    passed to the model during inference. Currently supports reasoning effort
    configuration, but can be extended with additional settings in the future.

    Attributes:
        reasoning_effort: Controls how much reasoning effort the model should apply when
            generating responses. Supported for tokenizer >= v15 and not supported for earlier versions.
    """

    reasoning_effort: ReasoningEffort | None = None

    @staticmethod
    def none() -> "ModelSettings":
        r"""Create a ModelSettings instance with default (None) values."""
        return ModelSettings()


class ResponseFormat(MistralBase):
    r"""The format of the response.

    Attributes:
        type: The type of the response.

    Examples:
        >>> response_format = ResponseFormat(type=ResponseFormats.text)
    """

    type: ResponseFormats = ResponseFormats.text


class ChatCompletionRequest(BaseCompletionRequest, Generic[ChatMessageType]):
    r"""Request for a chat completion.

    Attributes:
        model: The model to use for the chat completion.
        messages: The messages to use for the chat completion.
        response_format: The format of the response.
        tools: The tools to use for the chat completion.
        tool_choice: The tool choice to use for the chat completion.
        truncate_for_context_length: Whether to truncate the messages for the context length.
        continue_final_message: Whether to continue the final message.
        reasoning_effort: Controls how much reasoning effort the model should apply.

    Examples:
        >>> from mistral_common.protocol.instruct.messages import UserMessage, AssistantMessage
        >>> from mistral_common.protocol.instruct.tool_calls import ToolTypes, Function
        >>> request = ChatCompletionRequest(
        ...     messages=[
        ...         UserMessage(content="Hello!"),
        ...         AssistantMessage(content="Hi! How can I help you?"),
        ...     ],
        ...     response_format=ResponseFormat(type=ResponseFormats.text),
        ...     tools=[Tool(type=ToolTypes.function, function=Function(name="get_weather", parameters={}))],
        ...     tool_choice=ToolChoiceEnum.auto,
        ...     truncate_for_context_length=True,
        ... )
    """

    model: str | None = None
    messages: list[ChatMessageType]
    response_format: ResponseFormat = Field(default_factory=ResponseFormat)
    tools: list[Tool] | None = None
    tool_choice: ToolChoice = ToolChoiceEnum.auto
    truncate_for_context_length: bool = False
    continue_final_message: bool = False
    reasoning_effort: ReasoningEffort | None = None

    def to_openai(
        self,
        reasoning_field_format: ReasoningFieldFormat | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        r"""Convert the request messages and tools into the OpenAI format.

        Args:
            reasoning_field_format: Format for converting thinking chunks in assistant messages.
                See `AssistantMessage.to_openai` for details.
            kwargs: Additional parameters to be added to the request.

        Returns:
            The request in the OpenAI format.

        Examples:
            >>> from mistral_common.protocol.instruct.messages import UserMessage
            >>> from mistral_common.protocol.instruct.tool_calls import Tool, Function
            >>> request = ChatCompletionRequest(messages=[UserMessage(content="Hello, how are you?")], temperature=0.15)
            >>> request.to_openai(stream=True)
            {'temperature': 0.15, 'top_p': 1.0, 'response_format': {'type': 'text'}, 'continue_final_message': False, 'messages': [{'role': 'user', 'content': 'Hello, how are you?'}], 'tool_choice': 'auto', 'stream': True}
            >>> request = ChatCompletionRequest(messages=[UserMessage(content="Hello, how are you?")], tools=[
            ...     Tool(function=Function(
            ...         name="get_current_weather",
            ...         description="Get the current weather in a given location",
            ...         parameters={
            ...             "type": "object",
            ...             "properties": {
            ...                 "location": {
            ...                     "type": "string",
            ...                     "description": "The city and state, e.g. San Francisco, CA",
            ...                 },
            ...                 "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
            ...             },
            ...             "required": ["location"],
            ...         },
            ...     ),
            ... )])
            >>> request.to_openai()
            {'temperature': 0.7, 'top_p': 1.0, 'response_format': {'type': 'text'}, 'continue_final_message': False, 'messages': [{'role': 'user', 'content': 'Hello, how are you?'}], 'tools': [{'type': 'function', 'function': {'name': 'get_current_weather', 'description': 'Get the current weather in a given location', 'parameters': {'type': 'object', 'properties': {'location': {'type': 'string', 'description': 'The city and state, e.g. San Francisco, CA'}, 'unit': {'type': 'string', 'enum': ['celsius', 'fahrenheit']}}, 'required': ['location']}, 'strict': False}}], 'tool_choice': 'auto'}
        """  # noqa: E501

        # Handle messages and tools separately.
        openai_request: dict[str, Any] = self.model_dump(
            exclude={"messages", "tools", "truncate_for_context_length", "tool_choice"}, exclude_none=True
        )

        # Rename random_seed to seed.
        seed = openai_request.pop("random_seed", None)
        if seed is not None:
            openai_request["seed"] = seed

        if self.truncate_for_context_length:
            raise NotImplementedError("Truncating for context length is not implemented for OpenAI requests.")

        for kwarg in kwargs:
            if kwarg in openai_request:
                raise ValueError(f"Duplicate keyword argument: {kwarg}")
            elif kwarg in ChatCompletionRequest.model_fields:
                raise ValueError(f"Keyword argument {kwarg} is already set in the request.")

        openai_messages = []
        for message in self.messages:
            if isinstance(message, AssistantMessage):
                openai_messages.append(message.to_openai(reasoning_field_format=reasoning_field_format))
            else:
                openai_messages.append(message.to_openai())

        openai_request["messages"] = openai_messages
        if self.tools is not None:
            openai_request["tools"] = [tool.to_openai() for tool in self.tools]

        openai_tool_choice: str | dict[str, Any]
        match self.tool_choice:
            case ToolChoiceEnum.auto | ToolChoiceEnum.none:
                openai_tool_choice = self.tool_choice
            case ToolChoiceEnum.required | ToolChoiceEnum.any:
                openai_tool_choice = ToolChoiceEnum.required.value
            case _:
                openai_tool_choice = self.tool_choice.model_dump()

        openai_request["tool_choice"] = openai_tool_choice

        openai_request.update(kwargs)

        return openai_request

    @classmethod
    def from_openai(
        cls,
        messages: list[dict[str, str | list[dict[str, str | dict[str, Any]]]]],
        tools: list[dict[str, Any]] | None = None,
        continue_final_message: bool = False,
        **kwargs: Any,
    ) -> "ChatCompletionRequest":
        r"""Create a chat completion request from the OpenAI format.

        Args:
            messages: The messages in the OpenAI format.
            tools: The tools in the OpenAI format.
            continue_final_message: Whether to continue the final message.
            **kwargs: Additional keyword arguments to pass to the constructor. These should be the same as the fields
                of the request class or the OpenAI API equivalent.


        Returns:
            The chat completion request.
        """
        if "seed" in kwargs and "random_seed" in kwargs:
            raise ValueError("Cannot specify both `seed` and `random_seed`.")

        random_seed = kwargs.pop("seed", None)
        if random_seed is None:
            random_seed = kwargs.pop("random_seed", None)

        filtered_kwargs = cls._filter_cls_fields(kwargs)

        converted_messages: list[ChatMessage] = convert_openai_messages(messages)

        converted_tools = convert_openai_tools(tools) if tools is not None else None

        return cls(
            messages=converted_messages,  # type: ignore[arg-type]
            tools=converted_tools,
            random_seed=random_seed,
            continue_final_message=continue_final_message,
            **filtered_kwargs,
        )


class InstructRequest(MistralBase, Generic[ChatMessageType, ToolType]):
    r"""A valid Instruct request to be tokenized.

    Note:
        This class is intended for internal use only. External users should use `ChatCompletionRequest` to build
        requests and to convert to and from the OpenAI format.

    Attributes:
        messages: The history of the conversation.
        system_prompt: The system prompt to be used for the conversation.
        available_tools: The tools available to the assistant.
        truncate_at_max_tokens: The maximum number of tokens to truncate the conversation at.
        continue_final_message: Whether to continue the final message.
        settings: Model configuration settings for the request.

    Examples:
        >>> from mistral_common.protocol.instruct.messages import UserMessage, SystemMessage
        >>> request = InstructRequest(
        ...     messages=[UserMessage(content="Hello, how are you?")], system_prompt="You are a helpful assistant."
        ... )
    """

    messages: list[ChatMessageType]
    system_prompt: str | None = None
    available_tools: list[ToolType] | None = None
    truncate_at_max_tokens: int | None = None
    continue_final_message: bool = False
    settings: ModelSettings = Field(default_factory=ModelSettings.none)



from typing import Any

import pytest

from mistral_common.integrations.chat_templates.template_generator import build_chat_template
from mistral_common.protocol.instruct.request import ChatCompletionRequest
from mistral_common.protocol.instruct.validator import ValidationMode
from mistral_common.tokens.tokenizers.base import TokenizerVersion
from tests.integrations.chat_templates.conftest import ALL_CONFIGS, _config_id
from tests.integrations.chat_templates.fixtures_data import _get_conversations
from tests.integrations.chat_templates.helpers import (
    TestConfig,
    _load_golden_template,
    _make_config,
    render_template,
)


def _request_to_render_args(request: ChatCompletionRequest) -> dict[str, Any]:
    r"""Convert a ChatCompletionRequest to render_template kwargs.

    Enriches tool messages with the `name` field resolved from the preceding
    assistant's `tool_calls`, which the v2 Jinja template requires but
    `ToolMessage.to_openai()` omits.
    """
    openai = request.to_openai()
    messages = openai["messages"]

    # Build a mapping from tool_call_id -> function name so tool messages
    # can carry the `name` field expected by v2 templates.
    tool_call_names: dict[str, str] = {}
    for msg in messages:
        for tc in msg.get("tool_calls", []):
            tc_id = tc.get("id")
            fn_name = tc.get("function", {}).get("name")
            if tc_id and fn_name:
                tool_call_names[tc_id] = fn_name

    for msg in messages:
        if msg.get("role") == "tool" and "name" not in msg:
            tc_id = msg.get("tool_call_id")
            if tc_id and tc_id in tool_call_names:
                msg["name"] = tool_call_names[tc_id]

    kwargs: dict[str, Any] = {
        "messages": messages,
    }
    if "tools" in openai and openai["tools"]:
        kwargs["tools"] = openai["tools"]
    if (reasoning := openai.get("reasoning_effort")) is not None:
        kwargs["reasoning_effort"] = reasoning
    return kwargs


@pytest.mark.parametrize(
    "config",
    [c for c in ALL_CONFIGS if not c.plain_think],
    ids=[_config_id(c) for c in ALL_CONFIGS if not c.plain_think],
)
def test_dynamic_matches_static(config: TestConfig) -> None:
    template_config = _make_config(config)
    dynamic = build_chat_template(template_config)
    static = _load_golden_template(template_config)
    assert dynamic == static


@pytest.mark.parametrize(
    "config",
    [c for c in ALL_CONFIGS if not c.plain_think],
    ids=[_config_id(c) for c in ALL_CONFIGS if not c.plain_think],
)
def test_dynamic_template_produces_same_output(config: TestConfig) -> None:
    template_config = _make_config(config)
    static_template = _load_golden_template(template_config)
    dynamic_template = build_chat_template(template_config)

    # Test with a simple conversation
    messages: list[Any] = [
        {"role": "user", "content": "Hello"},
    ]

    # Only add assistant message for training validation
    if config.version != TokenizerVersion.v1:
        messages.append({"role": "assistant", "content": "Hi there!"})

    static_output = render_template(static_template, messages)
    dynamic_output = render_template(dynamic_template, messages)

    assert static_output == dynamic_output, (
        f"Output mismatch for {config}\n\nStatic output: {static_output}\nDynamic output: {dynamic_output}"
    )


# Comprehensive functional tests — covers all configs except plain_think
# (plain think template parity is tested separately in test_unit.py::test_plain_think_static_dynamic_parity)
@pytest.mark.parametrize(
    "config",
    [c for c in ALL_CONFIGS if not c.plain_think],
    ids=[_config_id(c) for c in ALL_CONFIGS if not c.plain_think],
)
def test_dynamic_template_comprehensive(config: TestConfig) -> None:
    template_config = _make_config(config)
    static_template = _load_golden_template(template_config)
    dynamic_template = build_chat_template(template_config)

    for mode in (ValidationMode.finetuning, ValidationMode.test):
        conversations = _get_conversations(config.version, mode, config.image, config.audio, config.think)
        for idx, request in enumerate(conversations):
            render_args = _request_to_render_args(request)
            static_output = render_template(static_template, **render_args)
            dynamic_output = render_template(dynamic_template, **render_args)
            assert static_output == dynamic_output, (
                f"Output mismatch for {config}, mode={mode}, conversation={idx}\n\n"
                f"Static output: {static_output}\n"
                f"Dynamic output: {dynamic_output}"
            )



from typing import Any

from PIL import Image

from mistral_common.protocol.instruct.chunk import (
    ContentChunk,
    ImageChunk,
    ImageURLChunk,
    TextChunk,
    ThinkChunk,
)
from tests.fixtures.audio import get_dummy_audio_chunk, get_dummy_audio_url_chunk


def get_content_chunk(name: str) -> ContentChunk:
    r"""Return a single instance of the requested content chunk type.

    Args:
        name: One of "text", "image", "image_url", "audio", "audio_url" or "think".

    Returns:
        A content chunk of the requested type.
    """
    chunks: dict[str, ContentChunk] = {
        "text": TextChunk(text="hello"),
        "image": ImageChunk(image=Image.new("RGB", (4, 4), "red")),
        "image_url": ImageURLChunk(image_url="data:image/png;base64,iVBORw0"),
        "audio": get_dummy_audio_chunk(),
        "audio_url": get_dummy_audio_url_chunk(),
        "think": ThinkChunk(thinking="reasoning"),
    }
    return chunks[name]


def get_content_chunks(names: tuple[str, ...]) -> list[Any]:
    r"""Return a list of content chunks for the requested type names.

    The element type is intentionally `Any` so the chunks can be fed to any message role
    constructor (including deliberately invalid combinations) in tests, leaving the runtime
    Pydantic validation to enforce the actual rules.
    """
    return [get_content_chunk(name) for name in names]



from typing import Any

import pytest

from mistral_common.integrations.chat_templates.chat_templates import generate_chat_template
from mistral_common.integrations.chat_templates.template_generator import (
    TemplateConfig,
    _render_content_call,
    build_chat_template,
)
from mistral_common.tokens.tokenizers.base import TokenizerVersion
from tests.integrations.chat_templates.helpers import TestConfig, _load_golden_template, _make_config, render_template


class TestGenerateChatTemplateAPI:
    @pytest.fixture()
    def v13_template(self) -> str:
        return generate_chat_template(
            spm=False,
            tokenizer_version=TokenizerVersion.v13,
            image_support=False,
            audio_support=False,
            thinking_support=False,
            default_system_prompt=None,
            plain_thinking_support=False,
            use_special_token_variables=True,
        )

    def test_basic_generation(self, v13_template: str) -> None:
        assert "{%- set default_system_message = '' %}" in v13_template
        assert "bos_token" in v13_template

    def test_default_system_prompt(self) -> None:
        template_with_prompt = generate_chat_template(
            spm=False,
            tokenizer_version=TokenizerVersion.v13,
            image_support=False,
            audio_support=False,
            thinking_support=False,
            default_system_prompt="You are a helpful assistant.",
            plain_thinking_support=False,
            use_special_token_variables=True,
        )
        assert "You are a helpful assistant." in template_with_prompt

    def test_basic_static_dynamic_parity(self, v13_template: str) -> None:
        config = _make_config(TestConfig(version=TokenizerVersion.v13))
        static_template = _load_golden_template(config)

        messages = [
            {"role": "user", "content": "Hello"},
            {"role": "assistant", "content": "Hi there!"},
        ]

        static_output = render_template(static_template, messages)
        dynamic_output = render_template(v13_template, messages)

        assert static_output == dynamic_output

    def test_empty_messages_raises(self) -> None:
        config = TemplateConfig(version=TokenizerVersion.v15)
        template = build_chat_template(config)
        with pytest.raises(Exception, match="list object has no element 0"):
            render_template(template, messages=[])

    def test_double_quotes_in_default_system_prompt(self) -> None:
        template = generate_chat_template(
            spm=False,
            tokenizer_version=TokenizerVersion.v7,
            image_support=False,
            audio_support=False,
            thinking_support=False,
            default_system_prompt='You are "the best" assistant.',
            plain_thinking_support=False,
            use_special_token_variables=True,
        )
        messages: list[dict[str, Any]] = [{"role": "user", "content": "Hello"}]
        result = render_template(template, messages=messages)
        assert 'You are "the best" assistant.' in result


class TestTokenVariables:
    @pytest.mark.parametrize(
        ("use_vars", "present", "absent"),
        [
            (True, ["bos_token", "eos_token"], ["{{- '<s>' }}", "{{- '</s>' }}"]),
            (False, ["'<s>'", "'</s>'"], ["bos_token", "eos_token"]),
        ],
    )
    def test_token_variable_content(self, use_vars: bool, present: list[str], absent: list[str]) -> None:
        config = TemplateConfig(version=TokenizerVersion.v7, use_special_token_variables=use_vars)
        template = build_chat_template(config)
        for expected in present:
            assert expected in template
        for unexpected in absent:
            assert unexpected not in template

    def test_use_special_token_variables_false_renders_correctly(self) -> None:
        config = TemplateConfig(version=TokenizerVersion.v15, use_special_token_variables=False)
        template = build_chat_template(config)
        messages: list[dict[str, Any]] = [
            {"role": "user", "content": "Hello"},
            {"role": "assistant", "content": "Hi there"},
        ]
        result = render_template(template, messages=messages)
        assert "<s>" in result
        assert "</s>" in result

    def test_use_special_token_variables_default_is_false(self) -> None:
        config = TemplateConfig(version=TokenizerVersion.v7)
        assert config.use_special_token_variables is False


class TestMessageContentValidation:
    @pytest.mark.parametrize("content", [[], None])
    @pytest.mark.parametrize("version", [TokenizerVersion.v3, TokenizerVersion.v7, TokenizerVersion.v15])
    def test_user_empty_content_renders_empty_turn(self, version: TokenizerVersion, content: list[Any] | None) -> None:
        template = generate_chat_template(
            spm=False,
            tokenizer_version=version,
            image_support=False,
            audio_support=False,
            thinking_support=False,
            default_system_prompt=None,
            plain_thinking_support=False,
            use_special_token_variables=True,
        )

        messages: list[dict[str, Any]] = [
            {"role": "user", "content": content},
        ]

        output = render_template(template, messages)
        assert "[INST]" in output
        assert "[/INST]" in output

    @pytest.mark.parametrize("version", [TokenizerVersion.v7, TokenizerVersion.v13, TokenizerVersion.v15])
    def test_assistant_empty_content_no_tool_calls_raises(self, version: TokenizerVersion) -> None:
        template = generate_chat_template(
            spm=False,
            tokenizer_version=version,
            image_support=False,
            audio_support=False,
            thinking_support=False,
            default_system_prompt=None,
            plain_thinking_support=False,
            use_special_token_variables=True,
        )

        messages: list[dict[str, Any]] = [
            {"role": "user", "content": "Hello"},
            {"role": "assistant", "content": ""},
        ]

        with pytest.raises(
            ValueError,
            match="Assistant message must have a string or a list of chunks in content or a list of tool calls",
        ):
            render_template(template, messages)

    def test_v1_empty_assistant_content_raises(self) -> None:
        template = generate_chat_template(
            spm=False,
            tokenizer_version=TokenizerVersion.v1,
            image_support=False,
            audio_support=False,
            thinking_support=False,
            default_system_prompt=None,
            plain_thinking_support=False,
            use_special_token_variables=True,
        )

        messages: list[dict[str, Any]] = [
            {"role": "user", "content": "Hello"},
            {"role": "assistant", "content": []},
        ]

        with pytest.raises(ValueError, match="Assistant message content must be non-empty"):
            render_template(template, messages)

    def test_v1_invalid_assistant_chunk_type_raises(self) -> None:
        template = generate_chat_template(
            spm=False,
            tokenizer_version=TokenizerVersion.v1,
            image_support=False,
            audio_support=False,
            thinking_support=False,
            default_system_prompt=None,
            plain_thinking_support=False,
            use_special_token_variables=True,
        )

        messages: list[dict[str, Any]] = [
            {"role": "user", "content": "Hello"},
            {"role": "assistant", "content": [{"type": "image", "url": "http://example.com/img.png"}]},
        ]

        with pytest.raises(ValueError, match="Only text chunks are supported in assistant message contents"):
            render_template(template, messages)


class TestRenderContentCall:
    r"""Unit tests for `_render_content_call` helper."""

    def test_minimal_text_only_config(self) -> None:
        """Text-only config emits only content and context_name."""
        config = TemplateConfig(version=TokenizerVersion.v7)
        result = _render_content_call(
            config=config,
            content_expr="message['content']",
            context_name="user message content",
            supported_types_desc="text",
            support_thinking=False,
            support_images=False,
            support_audio=False,
        )
        assert result == "render_content(content=message['content'], context_name='user message content')"

    @pytest.mark.parametrize(
        ("config", "call_kwargs", "present", "absent"),
        [
            pytest.param(
                TemplateConfig(version=TokenizerVersion.v13, thinking_support=True),
                {
                    "content_expr": "message['content']",
                    "context_name": "assistant message contents",
                    "supported_types_desc": "text and thinking",
                    "support_thinking": True,
                    "support_images": False,
                    "support_audio": False,
                },
                [
                    "content=message['content']",
                    "context_name='assistant message contents'",
                    "supported_types_desc='text and thinking'",
                    "support_thinking=true",
                ],
                ["support_images", "support_audio"],
                id="thinking_assistant",
            ),
            pytest.param(
                TemplateConfig(version=TokenizerVersion.v7, image_support=True),
                {
                    "content_expr": "user_content",
                    "context_name": "user message content",
                    "supported_types_desc": "text, image and image_url",
                    "support_thinking": False,
                    "support_images": True,
                    "support_audio": False,
                },
                ["support_images=true", "supported_types_desc='text, image and image_url'"],
                ["support_thinking", "support_audio"],
                id="image_user",
            ),
            pytest.param(
                TemplateConfig(version=TokenizerVersion.v7, audio_support=True),
                {
                    "content_expr": "message['content']",
                    "context_name": "user message content",
                    "supported_types_desc": "text, input_audio and audio_url",
                    "support_thinking": False,
                    "support_images": False,
                    "support_audio": True,
                },
                ["support_audio=true", "supported_types_desc='text, input_audio and audio_url'"],
                ["support_thinking", "support_images"],
                id="audio_user",
            ),
        ],
    )
    def test_feature_config_emits_correct_args(
        self,
        config: TemplateConfig,
        call_kwargs: dict,
        present: list[str],
        absent: list[str],
    ) -> None:
        """Feature-specific config emits the expected args and omits absent ones."""
        result = _render_content_call(config=config, **call_kwargs)
        for substring in present:
            assert substring in result
        for substring in absent:
            assert substring not in result

    def test_exact_thinking_image_assistant(self) -> None:
        """Full exact-string check for thinking+image config at assistant site."""
        config = TemplateConfig(version=TokenizerVersion.v13, thinking_support=True, image_support=True)
        result = _render_content_call(
            config=config,
            content_expr="message['content']",
            context_name="assistant message contents",
            supported_types_desc="text and thinking",
            support_thinking=True,
            support_images=False,
            support_audio=False,
        )
        assert result == (
            "render_content("
            "content=message['content'], "
            "context_name='assistant message contents', "
            "supported_types_desc='text and thinking', "
            "support_thinking=true, "
            "support_images=false)"
        )

    def test_exact_spm_image_user(self) -> None:
        """Full exact-string check for SPM+image config at user site."""
        config = TemplateConfig(version=TokenizerVersion.v3, spm=True, image_support=True)
        result = _render_content_call(
            config=config,
            content_expr="user_content",
            context_name="user message content",
            supported_types_desc="text, image and image_url",
            support_thinking=False,
            support_images=True,
            support_audio=False,
            initial_prev_img="not added_sp",
        )
        assert result == (
            "render_content("
            "content=user_content, "
            "context_name='user message content', "
            "supported_types_desc='text, image and image_url', "
            "support_images=true, "
            "initial_prev_img=not added_sp)"
        )

    def test_tool_ban_thinking_not_in_desc(self) -> None:
        """Tool call site with thinking config must have support_thinking=false and no 'thinking' in desc."""
        config = TemplateConfig(version=TokenizerVersion.v15, thinking_support=True)
        result = _render_content_call(
            config=config,
            content_expr="message['content']",
            context_name="tool message contents",
            supported_types_desc="text",
            support_thinking=False,
            support_images=False,
            support_audio=False,
        )
        assert "support_thinking=false" in result
        # The supported_types_desc must not mention thinking
        desc_value = result.split("supported_types_desc='")[1].split("'")[0]
        assert "thinking" not in desc_value

    def test_arg_order_matches_macro_declaration(self) -> None:
        """Args follow macro-declared order: content, context_name, supported_types_desc, support_thinking, support_images."""  # noqa: E501
        config = TemplateConfig(version=TokenizerVersion.v13, thinking_support=True, image_support=True)
        result = _render_content_call(
            config=config,
            content_expr="message['content']",
            context_name="assistant message contents",
            supported_types_desc="text and thinking",
            support_thinking=True,
            support_images=False,
            support_audio=False,
        )
        idx_content = result.index("content=")
        idx_context = result.index("context_name=")
        idx_desc = result.index("supported_types_desc=")
        idx_thinking = result.index("support_thinking=")
        idx_images = result.index("support_images=")
        assert idx_content < idx_context < idx_desc < idx_thinking < idx_images

    def test_image_config_emits_support_images(self) -> None:
        """Image config emits support_images, no support_thinking or support_audio."""
        config = TemplateConfig(version=TokenizerVersion.v7, image_support=True)
        result = _render_content_call(
            config=config,
            content_expr="user_content",
            context_name="user message content",
            supported_types_desc="text, image and image_url",
            support_thinking=False,
            support_images=True,
            support_audio=False,
        )
        assert "support_images=true" in result
        assert "support_thinking" not in result
        assert "support_audio" not in result

    def test_spm_image_emits_initial_prev_img(self) -> None:
        """SPM+image config emits initial_prev_img."""
        config = TemplateConfig(version=TokenizerVersion.v3, spm=True, image_support=True)
        result = _render_content_call(
            config=config,
            content_expr="user_content",
            context_name="user message content",
            supported_types_desc="text, image and image_url",
            support_thinking=False,
            support_images=True,
            support_audio=False,
            initial_prev_img="not added_sp",
        )
        assert "initial_prev_img=not added_sp" in result

    def test_non_spm_does_not_emit_initial_prev_img(self) -> None:
        """Non-SPM image config does not emit initial_prev_img."""
        config = TemplateConfig(version=TokenizerVersion.v7, image_support=True)
        result = _render_content_call(
            config=config,
            content_expr="user_content",
            context_name="user message content",
            supported_types_desc="text, image and image_url",
            support_thinking=False,
            support_images=True,
            support_audio=False,
        )
        assert "initial_prev_img" not in result

    def test_audio_config_emits_support_audio(self) -> None:
        """Audio config emits support_audio and supported_types_desc, no image or thinking."""
        config = TemplateConfig(version=TokenizerVersion.v7, audio_support=True)
        result = _render_content_call(
            config=config,
            content_expr="message['content']",
            context_name="user message content",
            supported_types_desc="text, input_audio and audio_url",
            support_thinking=False,
            support_images=False,
            support_audio=True,
        )
        assert "support_audio=true" in result
        assert "supported_types_desc='text, input_audio and audio_url'" in result
        assert "support_thinking" not in result
        assert "support_images" not in result



from pathlib import Path

from mistral_common.integrations.chat_templates.template_generator import (
    TemplateConfig,
)
from mistral_common.integrations.chat_templates.template_generator import (
    build_chat_template as _build_chat_template,
)
from mistral_common.tokens.tokenizers.base import SpecialTokens, TokenizerVersion
from mistral_common.tokens.tokenizers.mistral import MistralTokenizer
from mistral_common.tokens.tokenizers.sentencepiece import SentencePieceTokenizer


def generate_chat_template(
    spm: bool,
    tokenizer_version: TokenizerVersion,
    image_support: bool,
    audio_support: bool,
    thinking_support: bool,
    default_system_prompt: str | None,
    plain_thinking_support: bool,
    use_special_token_variables: bool,
) -> str:
    r"""Generate a chat template based on configuration.

    Programmatically generates a Jinja2 chat template string that formats
    conversation messages for Mistral models. The generated template handles
    message roles, special tokens, tool calls, and multimodal content.

    Args:
        spm: Whether to use SentencePiece tokenizer.
        tokenizer_version: The tokenizer version.
        image_support: Whether to support image chunks.
        audio_support: Whether to support audio chunks.
        thinking_support: Whether to support thinking chunks with special tokens.
        default_system_prompt: Optional default system prompt to embed.
        plain_thinking_support: Whether to support thinking chunks with plain
            `<think>`/`</think>` text tags. Only available for v11.
            Mutually exclusive with `thinking_support`.
        use_special_token_variables: Whether to emit BOS/EOS as Jinja variable
            references (`bos_token`/`eos_token`) or as literal values.

    Returns:
        The generated Jinja2 template as a string.
    """
    config = TemplateConfig(
        version=tokenizer_version,
        spm=spm,
        image_support=image_support,
        audio_support=audio_support,
        thinking_support=thinking_support,
        plain_thinking_support=plain_thinking_support,
        use_special_token_variables=use_special_token_variables,
    )
    template = _build_chat_template(config)

    if default_system_prompt is not None:
        # Thinking chunks are not a concern here because default system prompts
        # are plain text strings, not structured content with chunk types.
        escaped_prompt = default_system_prompt.replace("\\", "\\\\").replace("'", "\\'")
        template = template.replace(
            "{%- set default_system_message = '' %}",
            "{%- set default_system_message = '" + escaped_prompt + "' %}",
        )

    return template


def convert_tokenizer_to_chat_template(
    tokenizer_file: str | Path,
    system_prompt: str | None = None,
    use_special_token_variables: bool = True,
) -> str:
    r"""Load a tokenizer file and auto-detect its capabilities to generate a matching chat template.

    Loads the tokenizer via `MistralTokenizer.from_file`, inspects the resulting
    instruct tokenizer to determine version, backend (SentencePiece vs Tekken),
    and supported modalities, then delegates to `generate_chat_template` with the
    detected flags.

    The `plain_thinking_support` flag is set heuristically: a v11 tokenizer without
    an audio encoder uses plain `<think>`/`</think>` text tags instead of special
    `[THINK]`/`[/THINK]` tokens. When audio is present on a v11 tokenizer,
    `plain_thinking_support` is set to `False` because the two are mutually exclusive.
    `thinking_support` (special-token thinking) is detected by checking whether both
    `begin_think` and `end_think` special tokens are registered in the underlying
    tokenizer via its public `is_special` method.

    Args:
        tokenizer_file: Path to the tokenizer file (tekken JSON or SentencePiece `.model.vX`).
        system_prompt: Optional default system prompt to embed in the template.
            When not `None`, maps to `generate_chat_template`'s `default_system_prompt`.
        use_special_token_variables: Whether to emit BOS/EOS as Jinja variable
            references (`bos_token`/`eos_token`) or as literal string values.

    Returns:
        The generated Jinja2 chat template string matching the tokenizer's capabilities.

    Raises:
        TokenizerException: If the tokenizer file is not recognized or invalid.
    """
    mistral_tokenizer = MistralTokenizer.from_file(tokenizer_file)
    instruct_tokenizer = mistral_tokenizer.instruct_tokenizer
    tokenizer = instruct_tokenizer.tokenizer

    version = mistral_tokenizer.version
    spm = isinstance(tokenizer, SentencePieceTokenizer)
    image_support = instruct_tokenizer.image_encoder is not None
    audio_support = instruct_tokenizer.audio_encoder is not None
    thinking_support = tokenizer.is_special(SpecialTokens.begin_think.value) and tokenizer.is_special(
        SpecialTokens.end_think.value
    )
    plain_thinking_support = version == TokenizerVersion.v11 and not audio_support

    return generate_chat_template(
        spm=spm,
        tokenizer_version=version,
        image_support=image_support,
        audio_support=audio_support,
        thinking_support=thinking_support,
        default_system_prompt=system_prompt,
        plain_thinking_support=plain_thinking_support,
        use_special_token_variables=use_special_token_variables,
    )



r"""Chat template test configuration.

Provides package-level fixtures, parametrize config lists, and invalid-message
fixtures used across all chat template test layers.

Reusable helper functions live in `helpers.py`.
Test data constants (`REQUEST_*`) live in `fixtures_data.py`.
"""

from collections.abc import Generator
from typing import Any
from unittest.mock import patch

import pytest

from mistral_common.tokens.tokenizers.base import TokenizerVersion
from tests.integrations.chat_templates.helpers import _IMAGE, TestConfig

# All configurations for output comparison tests (including SPM).
ALL_CONFIGS: list[TestConfig] = [
    # Non-SPM
    TestConfig(version=TokenizerVersion.v1),
    TestConfig(version=TokenizerVersion.v2),
    TestConfig(version=TokenizerVersion.v3),
    TestConfig(version=TokenizerVersion.v3, image=True),
    TestConfig(version=TokenizerVersion.v7),
    TestConfig(version=TokenizerVersion.v7, image=True),
    TestConfig(version=TokenizerVersion.v7, audio=True),
    TestConfig(version=TokenizerVersion.v11),
    TestConfig(version=TokenizerVersion.v11, image=True),
    TestConfig(version=TokenizerVersion.v11, audio=True),
    TestConfig(version=TokenizerVersion.v13),
    TestConfig(version=TokenizerVersion.v13, image=True),
    TestConfig(version=TokenizerVersion.v13, audio=True),
    TestConfig(version=TokenizerVersion.v13, think=True),
    TestConfig(version=TokenizerVersion.v13, image=True, think=True),
    TestConfig(version=TokenizerVersion.v15),
    TestConfig(version=TokenizerVersion.v15, image=True),
    TestConfig(version=TokenizerVersion.v15, audio=True),
    TestConfig(version=TokenizerVersion.v15, think=True),
    TestConfig(version=TokenizerVersion.v15, image=True, think=True),
    # SPM
    TestConfig(version=TokenizerVersion.v1, spm=True),
    TestConfig(version=TokenizerVersion.v2, spm=True),
    TestConfig(version=TokenizerVersion.v3, spm=True),
    TestConfig(version=TokenizerVersion.v3, spm=True, image=True),
    TestConfig(version=TokenizerVersion.v7, spm=True),
    TestConfig(version=TokenizerVersion.v7, spm=True, image=True),
    # Plain thinking (v11 only)
    TestConfig(version=TokenizerVersion.v11, plain_think=True),
    TestConfig(version=TokenizerVersion.v11, image=True, plain_think=True),
]

# Parametrization configs for transformers tests (excludes plain_think).
ALL_TRANSFORMERS_CONFIGS: list[TestConfig] = [c for c in ALL_CONFIGS if not c.plain_think]


@pytest.fixture(scope="module")
def invalid_sp_think() -> dict[str, Any]:
    r"""Invalid system message containing a think chunk."""
    return {
        "messages": [
            {
                "role": "system",
                "content": [
                    {"type": "text", "text": "Hello"},
                    {"type": "think", "thinking": "Hello"},
                ],
            }
        ]
    }


@pytest.fixture(scope="module")
def invalid_sp_random() -> dict[str, Any]:
    r"""Invalid system message containing an unknown chunk type."""
    return {
        "messages": [
            {
                "role": "system",
                "content": [
                    {"type": "text", "text": "Hello"},
                    {"type": "random", "random": "Hello"},
                ],
            }
        ]
    }


@pytest.fixture(scope="module")
def invalid_assistant_think() -> dict[str, Any]:
    r"""Invalid assistant message containing a think chunk."""
    return {
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "Hello"},
                ],
            },
            {
                "role": "assistant",
                "content": [
                    {"type": "text", "text": "Hello"},
                    {"type": "think", "thinking": "Hello"},
                ],
            },
        ]
    }


@pytest.fixture(scope="module")
def invalid_assistant_random() -> dict[str, Any]:
    r"""Invalid assistant message containing an unknown chunk type."""
    return {
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "Hello"},
                ],
            },
            {
                "role": "assistant",
                "content": [
                    {"type": "text", "text": "Hello"},
                    {"type": "random", "random": "Hello"},
                ],
            },
        ]
    }


@pytest.fixture(scope="module")
def invalid_user_image() -> dict[str, Any]:
    r"""Invalid user message containing an image chunk."""
    return {
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "Hello"},
                    {"type": "image", "image_url": "Hello"},
                ],
            }
        ]
    }


@pytest.fixture(scope="module")
def invalid_user_audio() -> dict[str, Any]:
    r"""Invalid user message containing an audio chunk."""
    return {
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "Hello"},
                    {"type": "audio", "audio_url": "Hello"},
                ],
            }
        ]
    }


@pytest.fixture(scope="module")
def invalid_user_random() -> dict[str, Any]:
    r"""Invalid user message containing an unknown chunk type."""
    return {
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "Hello"},
                    {"type": "random", "random": "Hello"},
                ],
            }
        ]
    }


def _config_id(c: TestConfig) -> str:
    r"""Generate a human-readable test ID for a config."""
    parts = [c.version.value]
    if c.spm:
        parts.append("spm")
    if c.image:
        parts.append("img")
    if c.audio:
        parts.append("aud")
    if c.think:
        parts.append("think")
    if c.plain_think:
        parts.append("plain_think")
    return "_".join(parts)


@pytest.fixture(autouse=True, scope="package")
def mock_download_image() -> Generator[None, None, None]:
    r"""Mock `download_image` to return a dummy image for all tests."""
    with patch("mistral_common.image.download_image") as mock_download:
        mock_download.return_value = _IMAGE
        with patch("mistral_common.tokens.tokenizers.image.download_image") as mock_download2:
            mock_download2.return_value = _IMAGE
            yield



from typing import Any

import pytest

from mistral_common.exceptions import (
    InvalidAssistantMessageException,
    InvalidMessageStructureException,
    InvalidRequestException,
    InvalidSystemPromptException,
    InvalidToolMessageException,
    InvalidUserMessageException,
)
from mistral_common.protocol.instruct.chunk import (
    AudioChunk,
    AudioURLChunk,
)
from mistral_common.protocol.instruct.messages import (
    AssistantMessage,
    SystemMessage,
    ToolMessage,
    UserMessage,
)
from mistral_common.protocol.instruct.request import (
    ChatCompletionRequest,
    ReasoningEffort,
)
from mistral_common.protocol.instruct.tool_calls import FunctionCall, ToolCall
from mistral_common.protocol.instruct.validator import (
    MistralRequestValidator,
    MistralRequestValidatorV3,
    MistralRequestValidatorV5,
    MistralRequestValidatorV13,
    MistralRequestValidatorV15,
    ValidationMode,
)
from tests.fixtures.audio import get_dummy_audio_chunk, get_dummy_audio_url_chunk
from tests.fixtures.chunks import get_content_chunks

_Messages = list[UserMessage | AssistantMessage | SystemMessage | ToolMessage]


def _user_convo(content: "str | list[Any]") -> _Messages:
    return [UserMessage(content=content)]


def _assistant_convo(content: "str | list[Any]") -> _Messages:
    return [UserMessage(content="hi"), AssistantMessage(content=content), UserMessage(content="next")]


def _system_convo(content: "str | list[Any]") -> _Messages:
    return [SystemMessage(content=content), UserMessage(content="hi")]


def _tool_convo(content: "str | list[Any]") -> _Messages:
    return [
        UserMessage(content="hi"),
        AssistantMessage(tool_calls=[ToolCall(function=FunctionCall(name="fn", arguments="{}"), id="test12345")]),
        ToolMessage(content=content, tool_call_id="test12345"),
    ]


@pytest.fixture(scope="module")
def audio_chunk() -> AudioChunk:
    return get_dummy_audio_chunk()


@pytest.fixture(scope="module")
def audio_url_chunk() -> AudioURLChunk:
    return get_dummy_audio_url_chunk()


@pytest.fixture(
    params=[
        MistralRequestValidator(ValidationMode.serving),
        MistralRequestValidatorV3(ValidationMode.serving),
        MistralRequestValidatorV13(ValidationMode.serving),
    ]
)
def validator(request: pytest.FixtureRequest) -> MistralRequestValidator:
    return request.param  # type: ignore


@pytest.fixture
def validator_base() -> MistralRequestValidator:
    return MistralRequestValidator(ValidationMode.serving)


@pytest.fixture
def validator_v3() -> MistralRequestValidatorV3:
    return MistralRequestValidatorV3(ValidationMode.serving)


@pytest.fixture
def validator_v5() -> MistralRequestValidatorV5:
    return MistralRequestValidatorV5(ValidationMode.serving)


@pytest.fixture(
    params=[
        MistralRequestValidatorV13(ValidationMode.serving),
    ]
)
def validator_v13(request: pytest.FixtureRequest) -> MistralRequestValidator:
    return request.param  # type: ignore


@pytest.fixture(
    params=[
        MistralRequestValidatorV15(ValidationMode.serving),
    ]
)
def validator_v15(request: pytest.FixtureRequest) -> MistralRequestValidator:
    return request.param  # type: ignore


class TestChatValidation:
    def test_multiple_system_messages_OK(self, validator: MistralRequestValidator) -> None:
        validator.validate_messages(
            messages=[
                SystemMessage(content="foo"),
                SystemMessage(content="foo"),
                SystemMessage(content="foo"),
                UserMessage(content="foo"),  # so we don't get an error for ending with a system message
            ],
            continue_final_message=False,
        )

    def test_system_user_messages_OK(self, validator: MistralRequestValidator) -> None:
        validator.validate_messages(
            messages=[
                SystemMessage(content="foo"),
                UserMessage(content="foo"),
            ],
            continue_final_message=False,
        )

    def test_user_system_messages_OK(self, validator: MistralRequestValidator) -> None:
        validator.validate_messages(
            messages=[
                UserMessage(content="foo"),
                SystemMessage(content="foo"),
                UserMessage(content="foo"),  # so we don't get an error for ending with a system message
            ],
            continue_final_message=False,
        )

    def test_user_user_messages_OK(self, validator: MistralRequestValidator) -> None:
        validator.validate_messages(
            messages=[
                UserMessage(content="foo"),
                UserMessage(content="foo"),  # so we don't get an error for ending with a system message
            ],
            continue_final_message=False,
        )

    def test_empty_messages(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(InvalidMessageStructureException, match=r"Conversation must have at least one message"):
            validator.validate_messages(
                messages=[],
                continue_final_message=False,
            )

    def test_starts_with_system_or_user(self, validator: MistralRequestValidator) -> None:
        """
        This is allowed in validation, we will add a blank message in normalization
        """
        validator.validate_messages(
            messages=[
                AssistantMessage(content="foo"),
                UserMessage(content="foo"),  # so we don't get an error for ending with a system message
            ],
            continue_final_message=False,
        )

    def test_ends_with_assistant(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(
            InvalidMessageStructureException,
            match=(
                r"Expected last role User or Tool \(or Assistant with prefix or continue_final_message set to "
                r"True\) for serving but got assistant"
            ),
        ):
            validator.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(content="foo"),
                ],
                continue_final_message=False,
            )

    def test_assistant_prefix(self, validator: MistralRequestValidator) -> None:
        validator.validate_messages(
            messages=[
                UserMessage(content="foo"),
                AssistantMessage(content="foo", prefix=True),
            ],
            continue_final_message=False,
        )
        with pytest.raises(
            InvalidAssistantMessageException,
            match=r"Assistant message with prefix True must be last message",
        ):
            validator.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(content="foo", prefix=True),
                    UserMessage(content="foo"),
                ],
                continue_final_message=False,
            )

    def test_continue_final_message(self, validator: MistralRequestValidator) -> None:
        validator.validate_messages(
            messages=[
                UserMessage(content="foo"),
                AssistantMessage(content="foo"),
            ],
            continue_final_message=True,
        )
        with pytest.raises(
            InvalidMessageStructureException,
            match=(
                r"Expected last role Assistant with prefix False for serving with continue_final_message set to True "
                r"but got user"
            ),
        ):
            validator.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(content="foo", prefix=True),
                    UserMessage(content="foo"),
                ],
                continue_final_message=True,
            )
        with pytest.raises(
            InvalidMessageStructureException,
            match=(
                r"Expected last role User or Tool \(or Assistant with prefix or continue_final_message set to True\)"
                r" for serving but got assistant"
            ),
        ):
            validator.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(content="foo"),
                ],
                continue_final_message=False,
            )
        with pytest.raises(
            InvalidMessageStructureException,
            match=(
                r"Expected last role Assistant with prefix False for serving with continue_final_message set to True "
                r"but got assistant"
            ),
        ):
            validator.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(content="foo", prefix=True),
                ],
                continue_final_message=True,
            )

    def test_user_tool_user(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(InvalidMessageStructureException, match=r"Unexpected role 'tool' after role 'user'"):
            validator.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    ToolMessage(content="foo"),
                    UserMessage(content="foo"),
                ],
                continue_final_message=False,
            )

    def test_model_none_serving_mode(self) -> None:
        validator = MistralRequestValidator[UserMessage, AssistantMessage, ToolMessage, SystemMessage](
            ValidationMode.serving
        )
        with pytest.raises(InvalidRequestException, match=r"Model name parameter is required for serving mode"):
            validator.validate_request(ChatCompletionRequest(messages=[UserMessage(content="foo")]))

    def test_model_none_test_mode(self) -> None:
        validator = MistralRequestValidator[UserMessage, AssistantMessage, ToolMessage, SystemMessage](
            ValidationMode.test
        )
        validator.validate_request(ChatCompletionRequest(messages=[UserMessage(content="foo")]))

    def test_tool_calls_OK(self, validator: MistralRequestValidator) -> None:
        validator.validate_messages(
            messages=[
                UserMessage(content="foo"),
                AssistantMessage(
                    tool_calls=[ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}"))]
                ),
                ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
            ],
            continue_final_message=False,
        )

    def test_not_enough_tool_messages(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(
            InvalidMessageStructureException, match=r"Not the same number of function calls and responses"
        ):
            validator.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(
                        tool_calls=[
                            ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}")),
                            ToolCall(id="999999999", function=FunctionCall(name="foo", arguments="{}")),
                        ]
                    ),
                    ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
                    # missing ToolMessage
                ],
                continue_final_message=False,
            )

    def test_too_many_tool_messages(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(
            InvalidMessageStructureException,
            match=r"Not the same number of function calls and responses|Unexpected tool call id",
        ):
            validator.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(
                        tool_calls=[ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}"))]
                    ),
                    ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
                    ToolMessage(name="foo", content="bar", tool_call_id="999999999"),  # too many tool messages
                ],
                continue_final_message=False,
            )

    def test_tool_then_user_ok(self, validator: MistralRequestValidator) -> None:
        validator.validate_messages(
            messages=[
                UserMessage(content="foo"),
                AssistantMessage(
                    tool_calls=[ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}"))]
                ),
                ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
                UserMessage(content="continue with this context"),
            ],
            continue_final_message=False,
        )

    def test_tool_then_user_full_loop(self, validator: MistralRequestValidator) -> None:
        validator.validate_messages(
            messages=[
                UserMessage(content="search for info"),
                AssistantMessage(
                    tool_calls=[
                        ToolCall(id="aaaaaaaaa", function=FunctionCall(name="search", arguments="{}")),
                        ToolCall(id="bbbbbbbbb", function=FunctionCall(name="search", arguments="{}")),
                    ]
                ),
                ToolMessage(name="search", content="result1", tool_call_id="aaaaaaaaa"),
                ToolMessage(name="search", content="result2", tool_call_id="bbbbbbbbb"),
                UserMessage(content="now refine"),
                AssistantMessage(
                    tool_calls=[ToolCall(id="ccccccccc", function=FunctionCall(name="search", arguments="{}"))]
                ),
                ToolMessage(name="search", content="result3", tool_call_id="ccccccccc"),
                AssistantMessage(content="here is the final answer"),
                UserMessage(content="thanks"),
            ],
            continue_final_message=False,
        )

    def test_partial_tool_results_then_user_fails(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(
            InvalidMessageStructureException, match=r"Not the same number of function calls and responses"
        ):
            validator.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(
                        tool_calls=[
                            ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}")),
                            ToolCall(id="999999999", function=FunctionCall(name="bar", arguments="{}")),
                        ]
                    ),
                    ToolMessage(name="foo", content="result1", tool_call_id="123456789"),
                    UserMessage(content="stop, only need the first result"),
                ],
                continue_final_message=False,
            )

    def test_build_settings_raises_error(self, validator: MistralRequestValidator) -> None:
        request = ChatCompletionRequest(messages=[UserMessage(content="Hello")], reasoning_effort=ReasoningEffort.none)

        with pytest.raises(InvalidRequestException, match="reasoning_effort='none' is not supported for this model"):
            validator._validate_model_settings(request)

    def test_allows_text_content_chunks(self, validator_base: MistralRequestValidator) -> None:
        validator_base.validate_messages(_user_convo(get_content_chunks(("text",))), continue_final_message=False)
        validator_base.validate_messages(_assistant_convo(get_content_chunks(("text",))), continue_final_message=False)
        validator_base.validate_messages(_system_convo(get_content_chunks(("text",))), continue_final_message=False)
        validator_base.validate_messages(_tool_convo(get_content_chunks(("text",))), continue_final_message=False)

    def test_rejects_non_text_in_user(self, validator_base: MistralRequestValidator) -> None:
        for name in ("image", "image_url", "audio", "audio_url"):
            with pytest.raises(InvalidUserMessageException, match="Unexpected content chunk types in user message"):
                validator_base.validate_messages(_user_convo(get_content_chunks((name,))), continue_final_message=False)

    def test_rejects_non_text_in_assistant(self, validator_base: MistralRequestValidator) -> None:
        for name in ("think",):
            with pytest.raises(
                InvalidAssistantMessageException, match="Unexpected content chunk types in assistant message"
            ):
                validator_base.validate_messages(
                    _assistant_convo(get_content_chunks((name,))), continue_final_message=False
                )

    def test_rejects_non_text_in_system(self, validator_base: MistralRequestValidator) -> None:
        for name in ("audio", "think"):
            with pytest.raises(InvalidSystemPromptException, match="Unexpected content chunk types in system message"):
                validator_base.validate_messages(
                    _system_convo(get_content_chunks((name,))), continue_final_message=False
                )

    def test_rejects_non_text_in_tool(self, validator_base: MistralRequestValidator) -> None:
        # "think" is rejected at the Pydantic model level (ToolMessage construction), not by the validator.
        for name in ("image", "image_url", "audio", "audio_url"):
            with pytest.raises(InvalidToolMessageException, match="Unexpected content chunk types in tool message"):
                validator_base.validate_messages(_tool_convo(get_content_chunks((name,))), continue_final_message=False)

    def test_reports_sorted_unique_invalid_chunk_types(self, validator_base: MistralRequestValidator) -> None:
        content = get_content_chunks(("audio", "image_url"))
        with pytest.raises(
            InvalidUserMessageException,
            match=r"Unexpected content chunk types in user message: \['AudioChunk', 'ImageURLChunk'\]",
        ):
            validator_base.validate_messages(_user_convo(content), continue_final_message=False)


class TestChatValidationV3:
    def test_allows_text_and_image_in_user(self, validator_v3: MistralRequestValidatorV3) -> None:
        content = get_content_chunks(("text", "image", "image_url"))
        validator_v3.validate_messages(_user_convo(content), continue_final_message=False)

    def test_rejects_audio_in_user(self, validator_v3: MistralRequestValidatorV3) -> None:
        for name in ("audio", "audio_url"):
            with pytest.raises(InvalidUserMessageException, match="Unexpected content chunk types in user message"):
                validator_v3.validate_messages(_user_convo(get_content_chunks((name,))), continue_final_message=False)


class TestChatValidationV5:
    @pytest.mark.parametrize("audio_fixture", ["audio_chunk", "audio_url_chunk"])
    def test_audio_with_system_prompt_raises_error(
        self, validator_v5: MistralRequestValidatorV5, audio_fixture: str, request: pytest.FixtureRequest
    ) -> None:
        audio_chunk: AudioChunk | AudioURLChunk = request.getfixturevalue(audio_fixture)
        with pytest.raises(
            ValueError,
            match=(
                r"Found system messages at indexes \[0\] and audio chunks in messages at indexes \[1\]\. This is not "
                r"allowed prior to the tokenizer version 13\."
            ),
        ):
            validator_v5.validate_messages(
                messages=[
                    SystemMessage(content="This is a system prompt"),
                    UserMessage(content=[audio_chunk]),
                ],
                continue_final_message=False,
            )

    @pytest.mark.parametrize("audio_fixture", ["audio_chunk", "audio_url_chunk"])
    def test_audio_without_system_prompt_ok(
        self, validator_v5: MistralRequestValidatorV5, audio_fixture: str, request: pytest.FixtureRequest
    ) -> None:
        audio_chunk: AudioChunk | AudioURLChunk = request.getfixturevalue(audio_fixture)
        validator_v5.validate_messages(
            messages=[
                UserMessage(content=[audio_chunk]),
                UserMessage(content="User message after audio"),
            ],
            continue_final_message=False,
        )

    def test_system_prompt_without_audio_ok(self, validator_v5: MistralRequestValidatorV5) -> None:
        validator_v5.validate_messages(
            messages=[
                SystemMessage(content="This is a system prompt"),
                UserMessage(content="User message after system"),
            ],
            continue_final_message=False,
        )

    def test_build_settings_raises_error(self, validator: MistralRequestValidator) -> None:
        request = ChatCompletionRequest(messages=[UserMessage(content="Hello")], reasoning_effort=ReasoningEffort.none)

        with pytest.raises(InvalidRequestException, match="reasoning_effort='none' is not supported for this model"):
            validator._validate_model_settings(request)

    def test_allows_text_image_audio_in_user(self, validator_v5: MistralRequestValidatorV5) -> None:
        content = get_content_chunks(("text", "image", "image_url", "audio", "audio_url"))
        validator_v5.validate_messages(_user_convo(content), continue_final_message=False)

    def test_allows_text_audio_think_in_system(self, validator_v5: MistralRequestValidatorV5) -> None:
        content = get_content_chunks(("text", "audio", "think"))
        validator_v5.validate_messages(_system_convo(content), continue_final_message=False)

    def test_rejects_think_in_assistant(self, validator_v5: MistralRequestValidatorV5) -> None:
        for name in ("think",):
            with pytest.raises(
                InvalidAssistantMessageException, match="Unexpected content chunk types in assistant message"
            ):
                validator_v5.validate_messages(
                    _assistant_convo(get_content_chunks((name,))), continue_final_message=False
                )


class TestChatValidationV13:
    def test_right_number_results_invalid_id(self, validator_v13: MistralRequestValidatorV13) -> None:
        with pytest.raises(
            InvalidMessageStructureException,
            match=r"Unexpected tool call id",
        ):
            validator_v13.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(
                        tool_calls=[ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}"))]
                    ),
                    ToolMessage(name="foo", content="bar", tool_call_id="999999999"),  # invalid id
                ],
                continue_final_message=False,
            )

    def test_extra_results(self, validator_v13: MistralRequestValidatorV13) -> None:
        with pytest.raises(
            InvalidMessageStructureException,
            match=r"Unexpected tool call id",
        ):
            validator_v13.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(
                        tool_calls=[ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}"))]
                    ),
                    ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
                    ToolMessage(name="foo", content="bar", tool_call_id="999999999"),  # extra results
                ],
                continue_final_message=False,
            )

    def test_build_settings_raises_error(self, validator: MistralRequestValidator) -> None:
        request = ChatCompletionRequest(messages=[UserMessage(content="Hello")], reasoning_effort=ReasoningEffort.none)

        with pytest.raises(InvalidRequestException, match="reasoning_effort='none' is not supported for this model"):
            validator._validate_model_settings(request)

    def test_parallel_call_missing_results(self, validator_v13: MistralRequestValidatorV13) -> None:
        with pytest.raises(
            InvalidMessageStructureException,
            match=r"Not the same number of function calls and responses",
        ):
            validator_v13.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(
                        tool_calls=[
                            ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}")),
                            ToolCall(id="999999999", function=FunctionCall(name="foo", arguments="{}")),
                        ]
                    ),
                    ToolMessage(name="foo", content="bar", tool_call_id="123456789"),  # missing ToolMessage
                ],
                continue_final_message=False,
            )

    def allow_tool_results_wrong_order(self, validator_v13: MistralRequestValidatorV13) -> None:
        validator_v13.validate_messages(
            messages=[
                UserMessage(content="foo"),
                AssistantMessage(
                    tool_calls=[
                        ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="999999999", function=FunctionCall(name="foo", arguments="{}")),
                    ]
                ),
                ToolMessage(name="foo", content="bar", tool_call_id="999999999"),  # invalid order
                ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
            ],
            continue_final_message=False,
        )

    def test_tool_call_duplicate_ids_same_assistant_messages(self, validator_v13: MistralRequestValidatorV13) -> None:
        with pytest.raises(
            InvalidMessageStructureException,
            match=r"Duplicate",
        ):
            validator_v13.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(
                        tool_calls=[
                            ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}")),
                            ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}")),
                        ]
                    ),
                    ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
                ],
                continue_final_message=False,
            )

    def test_tool_call_duplicate_ids_different_assistant_messages(
        self, validator_v13: MistralRequestValidatorV13
    ) -> None:
        validator_v13.validate_messages(
            messages=[
                UserMessage(content="foo"),
                AssistantMessage(
                    tool_calls=[
                        ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}")),
                    ]
                ),
                ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
                AssistantMessage(
                    tool_calls=[
                        ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}")),
                    ]
                ),
                ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
            ],
            continue_final_message=False,
        )

    def test_multiple_assistant_messages_some_with_tool_calls(self, validator_v13: MistralRequestValidatorV13) -> None:
        validator_v13.validate_messages(
            messages=[
                UserMessage(content="foo"),
                AssistantMessage(
                    tool_calls=[
                        ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}")),
                    ]
                ),
                ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
                AssistantMessage(content="h"),
                UserMessage(content="foo"),
                AssistantMessage(
                    tool_calls=[
                        ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}")),
                    ]
                ),
                ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
                AssistantMessage(content="h"),
                UserMessage(content="foo"),
            ],
            continue_final_message=False,
        )

    def test_tool_then_user_ok_v13(self, validator_v13: MistralRequestValidatorV13) -> None:
        validator_v13.validate_messages(
            messages=[
                UserMessage(content="foo"),
                AssistantMessage(
                    tool_calls=[ToolCall(id="123456789", function=FunctionCall(name="foo", arguments="{}"))]
                ),
                ToolMessage(name="foo", content="bar", tool_call_id="123456789"),
                UserMessage(content="continue with this context"),
            ],
            continue_final_message=False,
        )

    def test_partial_tool_results_then_user_fails_v13(self, validator_v13: MistralRequestValidatorV13) -> None:
        with pytest.raises(
            InvalidMessageStructureException, match=r"Not the same number of function calls and responses"
        ):
            validator_v13.validate_messages(
                messages=[
                    UserMessage(content="foo"),
                    AssistantMessage(
                        tool_calls=[
                            ToolCall(id="aaaaaaaaa", function=FunctionCall(name="foo", arguments="{}")),
                            ToolCall(id="bbbbbbbbb", function=FunctionCall(name="bar", arguments="{}")),
                            ToolCall(id="ccccccccc", function=FunctionCall(name="baz", arguments="{}")),
                        ]
                    ),
                    ToolMessage(name="foo", content="result1", tool_call_id="aaaaaaaaa"),
                    UserMessage(content="stop, only need the first result"),
                ],
                continue_final_message=False,
            )

    def test_full_loop_with_user_after_tools_v13(self, validator_v13: MistralRequestValidatorV13) -> None:
        validator_v13.validate_messages(
            messages=[
                UserMessage(content="foo"),
                AssistantMessage(
                    tool_calls=[
                        ToolCall(id="aaaaaaaaa", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="bbbbbbbbb", function=FunctionCall(name="bar", arguments="{}")),
                    ]
                ),
                ToolMessage(name="foo", content="result1", tool_call_id="aaaaaaaaa"),
                ToolMessage(name="bar", content="result2", tool_call_id="bbbbbbbbb"),
                UserMessage(content="try something else"),
                AssistantMessage(
                    tool_calls=[ToolCall(id="ccccccccc", function=FunctionCall(name="baz", arguments="{}"))]
                ),
                ToolMessage(name="baz", content="result3", tool_call_id="ccccccccc"),
                AssistantMessage(content="final answer"),
                UserMessage(content="done"),
            ],
            continue_final_message=False,
        )

    @pytest.mark.parametrize("audio_fixture", ["audio_chunk", "audio_url_chunk"])
    def test_audio_with_system_prompt_raises_ok(
        self, validator_v13: MistralRequestValidatorV13, audio_fixture: str, request: pytest.FixtureRequest
    ) -> None:
        audio_chunk: AudioChunk | AudioURLChunk = request.getfixturevalue(audio_fixture)
        validator_v13.validate_messages(
            messages=[
                SystemMessage(content="This is a system prompt"),
                UserMessage(content=[audio_chunk]),
            ],
            continue_final_message=False,
        )

    def test_allows_text_and_think_in_assistant(self, validator_v13: MistralRequestValidatorV13) -> None:
        content = get_content_chunks(("text", "think"))
        validator_v13.validate_messages(_assistant_convo(content), continue_final_message=False)

    def test_rejects_non_text_in_tool(self, validator_v13: MistralRequestValidatorV13) -> None:
        # "think" is rejected at the Pydantic model level (ToolMessage construction), not by the validator.
        for name in ("image", "image_url", "audio", "audio_url"):
            with pytest.raises(InvalidToolMessageException, match="Unexpected content chunk types in tool message"):
                validator_v13.validate_messages(_tool_convo(get_content_chunks((name,))), continue_final_message=False)


class TestChatValidationV15:
    @pytest.mark.parametrize("reasoning_effort", [*list(ReasoningEffort), None])
    def test_build_settings_v15_reasoning_effort(
        self, reasoning_effort: ReasoningEffort | None, validator_v15: MistralRequestValidatorV15
    ) -> None:
        request = ChatCompletionRequest(messages=[UserMessage(content="Hello")], reasoning_effort=reasoning_effort)
        validator_v15._validate_model_settings(request)

    def test_allows_text_and_audio_in_system(self, validator_v15: MistralRequestValidatorV15) -> None:
        content = get_content_chunks(("text", "audio"))
        validator_v15.validate_messages(_system_convo(content), continue_final_message=False)

    def test_rejects_think_in_system(self, validator_v15: MistralRequestValidatorV15) -> None:
        for name in ("think",):
            with pytest.raises(InvalidSystemPromptException, match="Unexpected content chunk types in system message"):
                validator_v15.validate_messages(
                    _system_convo(get_content_chunks((name,))), continue_final_message=False
                )

    def test_allows_non_thinking_chunk_types_in_tool(self, validator_v15: MistralRequestValidatorV15) -> None:
        # ThinkChunk is rejected at the Pydantic model level before reaching the validator.
        content = get_content_chunks(("text", "image", "image_url", "audio", "audio_url"))
        validator_v15.validate_messages(_tool_convo(content), continue_final_message=False)



from pathlib import Path

import pytest

from mistral_common.exceptions import TokenizerException
from mistral_common.integrations.chat_templates.chat_templates import (
    convert_tokenizer_to_chat_template,
    generate_chat_template,
)
from mistral_common.tokens.tokenizers.base import TokenizerVersion
from tests.integrations.chat_templates.helpers import TestConfig, _build_spm_path, _build_tekken_json


class TestConvertTokenizerToChatTemplate:
    def test_tekken_v13_thinking(self, tmp_path: Path) -> None:
        config = TestConfig(version=TokenizerVersion.v13, think=True)
        path = _build_tekken_json(config=config, output_dir=tmp_path)
        result = convert_tokenizer_to_chat_template(tokenizer_file=path)
        expected = generate_chat_template(
            spm=False,
            tokenizer_version=TokenizerVersion.v13,
            image_support=False,
            audio_support=False,
            thinking_support=True,
            default_system_prompt=None,
            plain_thinking_support=False,
            use_special_token_variables=True,
        )
        assert result == expected
        assert "[THINK]" in result

    def test_tekken_v3_image(self, tmp_path: Path) -> None:
        config = TestConfig(version=TokenizerVersion.v3, image=True)
        path = _build_tekken_json(config=config, output_dir=tmp_path)
        result = convert_tokenizer_to_chat_template(tokenizer_file=path)
        expected = generate_chat_template(
            spm=False,
            tokenizer_version=TokenizerVersion.v3,
            image_support=True,
            audio_support=False,
            thinking_support=False,
            default_system_prompt=None,
            plain_thinking_support=False,
            use_special_token_variables=True,
        )
        assert result == expected

    def test_tekken_v7_audio(self, tmp_path: Path) -> None:
        config = TestConfig(version=TokenizerVersion.v7, audio=True)
        path = _build_tekken_json(config=config, output_dir=tmp_path)
        result = convert_tokenizer_to_chat_template(tokenizer_file=path)
        expected = generate_chat_template(
            spm=False,
            tokenizer_version=TokenizerVersion.v7,
            image_support=False,
            audio_support=True,
            thinking_support=False,
            default_system_prompt=None,
            plain_thinking_support=False,
            use_special_token_variables=True,
        )
        assert result == expected

    def test_tekken_v11_plain_thinking(self, tmp_path: Path) -> None:
        config = TestConfig(version=TokenizerVersion.v11)
        path = _build_tekken_json(config=config, output_dir=tmp_path)
        result = convert_tokenizer_to_chat_template(tokenizer_file=path)
        expected = generate_chat_template(
            spm=False,
            tokenizer_version=TokenizerVersion.v11,
            image_support=False,
            audio_support=False,
            thinking_support=False,
            default_system_prompt=None,
            plain_thinking_support=True,
            use_special_token_variables=True,
        )
        assert result == expected
        assert "<think>" in result
        assert "[THINK]" not in result

    def test_spm_v7(self, tmp_path: Path) -> None:
        config = TestConfig(version=TokenizerVersion.v7, spm=True)
        path = _build_spm_path(config=config, output_dir=tmp_path)
        result = convert_tokenizer_to_chat_template(tokenizer_file=path)
        expected = generate_chat_template(
            spm=True,
            tokenizer_version=TokenizerVersion.v7,
            image_support=False,
            audio_support=False,
            thinking_support=False,
            default_system_prompt=None,
            plain_thinking_support=False,
            use_special_token_variables=True,
        )
        assert result == expected

    def test_system_prompt_embedded(self, tmp_path: Path) -> None:
        config = TestConfig(version=TokenizerVersion.v7)
        path = _build_tekken_json(config=config, output_dir=tmp_path)
        result = convert_tokenizer_to_chat_template(
            tokenizer_file=path,
            system_prompt="You are helpful.",
        )
        assert "You are helpful." in result

    def test_use_special_token_variables_true(self, tmp_path: Path) -> None:
        config = TestConfig(version=TokenizerVersion.v7)
        path = _build_tekken_json(config=config, output_dir=tmp_path)
        result = convert_tokenizer_to_chat_template(
            tokenizer_file=path,
            use_special_token_variables=True,
        )
        assert "bos_token" in result
        assert "'<s>'" not in result

    def test_use_special_token_variables_false(self, tmp_path: Path) -> None:
        config = TestConfig(version=TokenizerVersion.v7)
        path = _build_tekken_json(config=config, output_dir=tmp_path)
        result = convert_tokenizer_to_chat_template(
            tokenizer_file=path,
            use_special_token_variables=False,
        )
        assert "'<s>'" in result
        assert "bos_token" not in result

    def test_tekken_v11_audio_no_plain_thinking(self, tmp_path: Path) -> None:
        config = TestConfig(version=TokenizerVersion.v11, audio=True)
        path = _build_tekken_json(config=config, output_dir=tmp_path)
        result = convert_tokenizer_to_chat_template(tokenizer_file=path)
        expected = generate_chat_template(
            spm=False,
            tokenizer_version=TokenizerVersion.v11,
            image_support=False,
            audio_support=True,
            thinking_support=False,
            default_system_prompt=None,
            plain_thinking_support=False,
            use_special_token_variables=True,
        )
        assert result == expected
        assert "[AUDIO]" in result
        assert "<think>" not in result

    def test_unrecognized_file_raises_tokenizer_exception(self, tmp_path: Path) -> None:
        invalid_path = tmp_path / "tokenizer.txt"
        invalid_path.write_text("not a tokenizer")
        with pytest.raises(TokenizerException, match="Unrecognized tokenizer file:"):
            convert_tokenizer_to_chat_template(tokenizer_file=invalid_path)



import subprocess
import sys
from pathlib import Path

import pytest

from mistral_common.tokens.tokenizers.base import TokenizerVersion
from tests.integrations.chat_templates.helpers import TestConfig, _build_tekken_json

SCRIPT_PATH = Path(__file__).parent.parent / "scripts" / "generate_chat_template.py"


@pytest.fixture(scope="session")
def tekken_think_v13_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    """Path to a v13 think tekken.json used for auto-detect CLI tests."""
    build_dir = tmp_path_factory.mktemp("tokenizer")
    config = TestConfig(version=TokenizerVersion.v13, think=True)
    return _build_tekken_json(config=config, output_dir=build_dir)


def test_cli_generates_template(tmp_path: Path) -> None:
    """CLI generates a template file with expected content."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [sys.executable, str(SCRIPT_PATH), "--version", "v15", "--saving_path", str(output_path)],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    assert output_path.exists()
    content = output_path.read_text()
    assert "bos_token" in content
    assert "[MODEL_SETTINGS]" in content


def test_cli_with_image_flag(tmp_path: Path) -> None:
    """CLI --image flag produces template with image support."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [sys.executable, str(SCRIPT_PATH), "--version", "v15", "--image", "--saving_path", str(output_path)],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    content = output_path.read_text()
    assert "image" in content.lower()


def test_cli_with_thinking_flag(tmp_path: Path) -> None:
    """CLI --thinking flag produces template with thinking support."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [sys.executable, str(SCRIPT_PATH), "--version", "v15", "--thinking", "--saving_path", str(output_path)],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    content = output_path.read_text()
    assert "[THINK]" in content


def test_cli_with_default_system_prompt(tmp_path: Path) -> None:
    """CLI --default_system_prompt embeds the prompt."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [
            sys.executable,
            str(SCRIPT_PATH),
            "--version",
            "v7",
            "--default_system_prompt",
            "You are helpful.",
            "--saving_path",
            str(output_path),
        ],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    content = output_path.read_text()
    assert "You are helpful." in content


def test_cli_invalid_version(tmp_path: Path) -> None:
    """CLI rejects invalid version."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [sys.executable, str(SCRIPT_PATH), "--version", "v99", "--saving_path", str(output_path)],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0


def test_cli_invalid_config(tmp_path: Path) -> None:
    """CLI rejects invalid configuration (e.g., image + audio)."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [sys.executable, str(SCRIPT_PATH), "--version", "v15", "--image", "--audio", "--saving_path", str(output_path)],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0


def test_cli_missing_version(tmp_path: Path) -> None:
    """CLI requires --version in manual mode and emits the specific error message."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [sys.executable, str(SCRIPT_PATH), "--saving_path", str(output_path)],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0
    assert "--version is required when --tokenizer_file is not provided" in result.stderr


def test_cli_with_audio_flag(tmp_path: Path) -> None:
    """CLI --audio flag produces template with audio support."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [sys.executable, str(SCRIPT_PATH), "--version", "v7", "--audio", "--saving_path", str(output_path)],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    content = output_path.read_text()
    assert "[AUDIO]" in content


def test_cli_with_plain_thinking_flag(tmp_path: Path) -> None:
    """CLI --plain_thinking flag produces template with plain text thinking."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [sys.executable, str(SCRIPT_PATH), "--version", "v11", "--plain_thinking", "--saving_path", str(output_path)],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    content = output_path.read_text()
    assert "<think>" in content
    assert "[THINK]" not in content


def test_cli_plain_thinking_invalid_version(tmp_path: Path) -> None:
    """CLI rejects --plain_thinking with non-v11 version."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [sys.executable, str(SCRIPT_PATH), "--version", "v15", "--plain_thinking", "--saving_path", str(output_path)],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0


def test_cli_spm_flag(tmp_path: Path) -> None:
    """CLI --spm flag produces SPM template."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [sys.executable, str(SCRIPT_PATH), "--version", "v7", "--spm", "--saving_path", str(output_path)],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    content = output_path.read_text()
    assert "bos_token" in content


def test_cli_spm_with_audio_invalid(tmp_path: Path) -> None:
    """CLI --spm --audio together fails validation."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [sys.executable, str(SCRIPT_PATH), "--version", "v7", "--spm", "--audio", "--saving_path", str(output_path)],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0


def test_cli_no_special_token_variables_flag(tmp_path: Path) -> None:
    """CLI --no_special_token_variables embeds literal BOS/EOS values."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [
            sys.executable,
            str(SCRIPT_PATH),
            "--version",
            "v7",
            "--no_special_token_variables",
            "--saving_path",
            str(output_path),
        ],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    content = output_path.read_text()
    assert "'<s>'" in content
    assert "'</s>'" in content
    assert "bos_token" not in content
    assert "eos_token" not in content


def test_autodetect_happy_path(tmp_path: Path, tekken_think_v13_path: Path) -> None:
    """Auto-detect mode with v13 think tokenizer generates template containing [THINK]."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [
            sys.executable,
            str(SCRIPT_PATH),
            "--tokenizer_file",
            str(tekken_think_v13_path),
            "--saving_path",
            str(output_path),
        ],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    assert output_path.exists()
    content = output_path.read_text()
    assert "[THINK]" in content


def test_autodetect_conflict_version(tmp_path: Path, tekken_think_v13_path: Path) -> None:
    """Auto-detect + --version exits non-zero with mutual-exclusion error."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [
            sys.executable,
            str(SCRIPT_PATH),
            "--tokenizer_file",
            str(tekken_think_v13_path),
            "--version",
            "v13",
            "--saving_path",
            str(output_path),
        ],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0
    assert "--tokenizer_file cannot be combined with manual capability flags" in result.stderr


def test_autodetect_conflict_capability_flag(tmp_path: Path, tekken_think_v13_path: Path) -> None:
    """Auto-detect + --image exits non-zero with mutual-exclusion error."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [
            sys.executable,
            str(SCRIPT_PATH),
            "--tokenizer_file",
            str(tekken_think_v13_path),
            "--image",
            "--saving_path",
            str(output_path),
        ],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0
    assert "--tokenizer_file cannot be combined with manual capability flags" in result.stderr


def test_autodetect_with_system_prompt(tmp_path: Path, tekken_think_v13_path: Path) -> None:
    """Auto-detect mode with --default_system_prompt embeds the prompt."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [
            sys.executable,
            str(SCRIPT_PATH),
            "--tokenizer_file",
            str(tekken_think_v13_path),
            "--default_system_prompt",
            "You are helpful.",
            "--saving_path",
            str(output_path),
        ],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    content = output_path.read_text()
    assert "You are helpful." in content


def test_autodetect_no_special_token_variables(tmp_path: Path, tekken_think_v13_path: Path) -> None:
    """Auto-detect + --no_special_token_variables embeds literal BOS/EOS values."""
    output_path = tmp_path / "template.jinja"
    result = subprocess.run(
        [
            sys.executable,
            str(SCRIPT_PATH),
            "--tokenizer_file",
            str(tekken_think_v13_path),
            "--no_special_token_variables",
            "--saving_path",
            str(output_path),
        ],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    content = output_path.read_text()
    assert "'<s>'" in content
    assert "bos_token" not in content



import pytest

from mistral_common.experimental.tools import (
    InvalidArgsToolCallError,
    InvalidToolCallError,
    _decode_tool_calls,
    _split_content_and_tool_calls,
    _split_integer_list_by_value,
)
from mistral_common.protocol.instruct.messages import (
    AssistantMessage,
    SystemMessage,
    ToolMessage,
    UserMessage,
)
from mistral_common.protocol.instruct.normalize import InstructRequestNormalizerV7, InstructRequestNormalizerV13
from mistral_common.protocol.instruct.request import InstructRequest
from mistral_common.protocol.instruct.tool_calls import FunctionCall, ToolCall
from mistral_common.protocol.instruct.validator import (
    MistralRequestValidatorV5,
    MistralRequestValidatorV13,
)
from mistral_common.tokens.tokenizers.base import TokenizerVersion
from mistral_common.tokens.tokenizers.instruct import InstructTokenizerV11, InstructTokenizerV13
from mistral_common.tokens.tokenizers.mistral import MistralTokenizer
from mistral_common.tokens.tokenizers.tekken import Tekkenizer
from tests.test_tekken import get_special_tokens, quick_vocab


def fixture_mistral_tokenizer_v11() -> MistralTokenizer:
    return MistralTokenizer(
        instruct_tokenizer=InstructTokenizerV11(
            Tekkenizer(
                quick_vocab(
                    [
                        b"Hello",
                        b",",
                        b" ",
                        b"world",
                        b"!",
                        b"How",
                        b"can",
                        b"I",
                        b"assist",
                        b"you",
                        b"today",
                        b"?",
                        b'"',
                        b"a",
                        b"b",
                        b"c",
                        b"d",
                        b"{",
                        b"}",
                        b":",
                    ]
                ),
                special_tokens=get_special_tokens(TokenizerVersion.v11),
                pattern=r".+",  # single token, whole string
                vocab_size=256 + 100,
                num_special_tokens=100,
                version=TokenizerVersion.v11,
            ),
        ),
        validator=MistralRequestValidatorV5(),
        request_normalizer=InstructRequestNormalizerV7(
            UserMessage, AssistantMessage, ToolMessage, SystemMessage, InstructRequest, None
        ),
    )


def fixture_mistral_tokenizer_v13() -> MistralTokenizer:
    return MistralTokenizer(
        instruct_tokenizer=InstructTokenizerV13(
            Tekkenizer(
                quick_vocab(
                    [
                        b"Hello",
                        b",",
                        b" ",
                        b"world",
                        b"!",
                        b"How",
                        b"can",
                        b"I",
                        b"assist",
                        b"you",
                        b"today",
                        b"?",
                        b'"',
                        b"a",
                        b"b",
                        b"c",
                        b"d",
                        b"{",
                        b"}",
                        b"1",
                        b"2",
                        b"call",
                        b"_",
                        b":",
                    ]
                ),
                special_tokens=get_special_tokens(TokenizerVersion.v13),
                pattern=r".+",  # single token, whole string
                vocab_size=256 + 100,
                num_special_tokens=100,
                version=TokenizerVersion.v13,
            )
        ),
        validator=MistralRequestValidatorV13(),
        request_normalizer=InstructRequestNormalizerV13(
            UserMessage, AssistantMessage, ToolMessage, SystemMessage, InstructRequest, None
        ),
    )


def test_find_content_tool_calls() -> None:
    # Test 1: No tool calls
    tokens = [1, 2, 3, 4, 5]
    assert _split_content_and_tool_calls(tokens, 6) == ([1, 2, 3, 4, 5], ())

    # Test 2: One tool call
    tokens = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    assert _split_content_and_tool_calls(tokens, 6) == ([1, 2, 3, 4, 5], ([6, 7, 8, 9, 10],))

    # Test 3: Multiple tool calls
    tokens = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 6, 11, 12, 13, 14]
    assert _split_content_and_tool_calls(tokens, 6) == ([1, 2, 3, 4, 5], ([6, 7, 8, 9, 10], [6, 11, 12, 13, 14]))

    # Test 4: No content
    tokens = [6, 7, 8, 9, 10]
    assert _split_content_and_tool_calls(tokens, 6) == ([], ([6, 7, 8, 9, 10],))


@pytest.mark.parametrize(
    "tokenizer",
    (
        MistralTokenizer.v2(),
        MistralTokenizer.v3(),
        MistralTokenizer.v7(),
        MistralTokenizer(
            instruct_tokenizer=InstructTokenizerV11(
                Tekkenizer(
                    quick_vocab(
                        [
                            b"Hello",
                            b",",
                            b" ",
                            b"world",
                            b"!",
                            b"How",
                            b"can",
                            b"I",
                            b"assist",
                            b"you",
                            b"today",
                            b"?",
                            b'"',
                            b"a",
                            b"b",
                            b"c",
                            b"d",
                            b"{",
                            b"}",
                            b":",
                        ]
                    ),
                    special_tokens=get_special_tokens(TokenizerVersion.v11),
                    pattern=r".+",  # single token, whole string
                    vocab_size=256 + 100,
                    num_special_tokens=100,
                    version=TokenizerVersion.v11,
                ),
            ),
            validator=MistralRequestValidatorV5(),
            request_normalizer=InstructRequestNormalizerV7(
                UserMessage, AssistantMessage, ToolMessage, SystemMessage, InstructRequest, None
            ),
        ),
        fixture_mistral_tokenizer_v11(),
        fixture_mistral_tokenizer_v13(),
    ),
)
@pytest.mark.parametrize(
    "tool_calls",
    (
        [ToolCall(id="call_1", function=FunctionCall(name="ab", arguments="{}"))],
        [
            ToolCall(id="call_1", function=FunctionCall(name="ab", arguments="{}")),
            ToolCall(id="call_2", function=FunctionCall(name="cd", arguments='{"a": "b", "c": "d"}')),
        ],
    ),
)
def test_decode_tool_calls(tokenizer: MistralTokenizer, tool_calls: list[ToolCall]) -> None:
    # Test 1:
    # valid tool calls
    encoded_tool_calls = tokenizer.instruct_tokenizer._encode_tool_calls_in_assistant_message(  # type: ignore[attr-defined]
        AssistantMessage(tool_calls=tool_calls)
    )

    splitted_tool_calls = _split_integer_list_by_value(
        encoded_tool_calls, tokenizer.instruct_tokenizer.tokenizer.get_special_token("[TOOL_CALLS]")
    )

    decoded_tool_calls = _decode_tool_calls(splitted_tool_calls, tokenizer.instruct_tokenizer.tokenizer)
    if tokenizer.instruct_tokenizer.tokenizer.version != TokenizerVersion.v11:
        assert len(decoded_tool_calls) == len(tool_calls)
        for decoded_tool_call, tool_call in zip(decoded_tool_calls, tool_calls):
            assert decoded_tool_call.model_dump(exclude={"id"}) == tool_call.model_dump(exclude={"id"})
    else:
        # v11 is the only version that encode the id
        assert decoded_tool_calls == tool_calls

    versions_inf_v11 = [
        TokenizerVersion.v2,
        TokenizerVersion.v3,
        TokenizerVersion.v7,
    ]

    # Test 2:
    # invalid tool calls (remove a JSON needed token from the args)
    if tokenizer.instruct_tokenizer.tokenizer.version in versions_inf_v11:
        arg_index = tokenizer.instruct_tokenizer.tokenizer.get_special_token("[TOOL_CALLS]")
    else:
        arg_index = tokenizer.instruct_tokenizer.tokenizer.get_special_token("[ARGS]")

    first_args_index = splitted_tool_calls[0].index(arg_index)
    splitted_tool_calls = (
        splitted_tool_calls[0][: first_args_index + 1]
        + splitted_tool_calls[0][first_args_index + 2 :],  # make the args invalid
        *splitted_tool_calls[1:],
    )

    with pytest.raises(
        InvalidToolCallError
        if tokenizer.instruct_tokenizer.tokenizer.version in versions_inf_v11
        else InvalidArgsToolCallError
    ):
        _decode_tool_calls(splitted_tool_calls, tokenizer.instruct_tokenizer.tokenizer)


def test_decode_tool_calls_v11_without_id() -> None:
    tokenizer = fixture_mistral_tokenizer_v11()
    tool_call = ToolCall(id="call_1", function=FunctionCall(name="ab", arguments="{}"))

    encoded_tool_call: list[int] = tokenizer.instruct_tokenizer._encode_tool_calls_in_assistant_message(  # type: ignore[attr-defined]
        AssistantMessage(tool_calls=[tool_call])
    )

    call_id_token = tokenizer.instruct_tokenizer.tokenizer.get_special_token("[CALL_ID]")
    args_token = tokenizer.instruct_tokenizer.tokenizer.get_special_token("[ARGS]")

    call_id_index = encoded_tool_call.index(call_id_token)
    args_token_index = encoded_tool_call.index(args_token)

    encoded_tool_call = encoded_tool_call[:call_id_index] + encoded_tool_call[args_token_index:]  # remove the id

    decoded_tool_call = _decode_tool_calls([encoded_tool_call], tokenizer.instruct_tokenizer.tokenizer)
    assert len(decoded_tool_call) == 1
    assert decoded_tool_call[0].model_dump(exclude={"id"}) == tool_call.model_dump(exclude={"id"})
